# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the vit implementation of apple/ml-cvnets.
https://github.com/apple/ml-cvnets/blob/84d992f413e52c0468f86d23196efd9dad885e6f/cvnets/models/classification/vit.py
"""
import argparse
from typing import Any, Dict, Optional, Tuple, Union, List

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ...op.base_metaformer import (
    ChannelMLP,
    Image2Sequence,
    MetaFormer,
    MetaFormerBlock,
    MetaFormerEncoder,
    MultiHeadAttention,
)
from ...op.custom import ConvLayer, SinusoidalPositionalEncoding
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['vit']
SUPPORTING_TASK = ['classification']


class ViTEmbeddings(nn.Module):
    def __init__(self, image_channels, patch_size, hidden_size, hidden_dropout_prob, use_cls_token=True, vocab_size=1000):
        super().__init__()
        
        image_channels = 3  # {RGB}
        
        kernel_size = patch_size
        if patch_size % 2 == 0:
            kernel_size += 1
        
        self.patch_emb = ConvLayer(
            in_channels=image_channels,
            out_channels=hidden_size,
            kernel_size=kernel_size,
            stride=patch_size,
            bias=True,
            use_norm=False,
            use_act=False,
        )
        self.flat = Image2Sequence(contiguous=True)
        
        self.cls_token = None
        if use_cls_token:
            self.cls_token = nn.Parameter(torch.zeros(size=(1, 1, hidden_size)))
            torch.nn.init.trunc_normal_(self.cls_token, std=0.02)
            
        self.pos_embed = SinusoidalPositionalEncoding(
                d_model=hidden_size,
                channels_last=True,
                max_len=vocab_size,
        )
        self.dropout = nn.Dropout(p=hidden_dropout_prob)
    
    def forward(self, x):
        B = x.shape[0]  # B x 3(={RGB}) x H x W

        patch_emb = self.patch_emb(x)  # B x C(=embed_dim) x H'(=patch_size) x W'(=patch_size)
        patch_emb = self.flat(patch_emb)  # B x H'*W' x C

        # add classification token
        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(B, -1, -1)  # B x 1 x C
            patch_emb = torch.cat((cls_tokens, patch_emb), dim=1)  # B x (H'*W' + 1) x C

        patch_emb = self.pos_embed(patch_emb)  # B x (H'*W' + 1) x C
        
        patch_emb = self.dropout(patch_emb)  # B x (H'*W' + 1) x C
        return patch_emb  # B x (H'*W' + 1) x C


class ViTBlock(MetaFormerBlock):
    def __init__(self, hidden_size, num_attention_heads, attention_dropout_prob, intermediate_size, hidden_dropout_prob, layer_norm_eps) -> None:
        super().__init__(hidden_size, layer_norm_eps)
        self.layernorm_before = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.layernorm_after = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.token_mixer = MultiHeadAttention(hidden_size, num_attention_heads,
                                                  attention_dropout_prob=attention_dropout_prob,
                                                  use_qkv_bias=True
                                                  )
        self.channel_mlp = ChannelMLP(hidden_size, intermediate_size, hidden_dropout_prob, hidden_activation_type='gelu')
    
    
class ViTEncoder(MetaFormerEncoder):
    def __init__(self, num_blocks, hidden_size, num_attention_heads, attention_dropout_prob, intermediate_size, hidden_dropout_prob, layer_norm_eps) -> None:
        super().__init__()
        self.blocks = nn.Sequential(
            *[ViTBlock(hidden_size, num_attention_heads, attention_dropout_prob, intermediate_size, hidden_dropout_prob, layer_norm_eps) for _ in range(num_blocks)]
        )


class VisionTransformer(MetaFormer):
    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'ViT is not supported on {self.task} task now.'
        #self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        patch_size = params.patch_size
        hidden_size = params.attention_channels
        num_blocks = params.num_blocks
        num_attention_heads = params.num_attention_heads
        attention_dropout_prob = params.attention_dropout_prob
        intermediate_size = params.ffn_intermediate_channels
        hidden_dropout_prob = params.ffn_dropout_prob
        use_cls_token = params.use_cls_token
        vocab_size = params.vocab_size

        # Fix as a constant
        layer_norm_eps = 1e-5

        hidden_sizes = hidden_size if isinstance(hidden_size, list) else [hidden_size] * num_blocks
        super().__init__(hidden_sizes)
        self.task = task
        self.intermediate_features = self.task in ['segmentation', 'detection']
        
        image_channels = 3
        self.patch_embed = ViTEmbeddings(image_channels, patch_size, hidden_sizes[-1], hidden_dropout_prob, use_cls_token=use_cls_token, vocab_size=vocab_size)
        self.encoder = ViTEncoder(num_blocks, hidden_sizes[-1], num_attention_heads, attention_dropout_prob, intermediate_size, hidden_dropout_prob, layer_norm_eps)
        self.norm = nn.LayerNorm(hidden_sizes[-1], eps=layer_norm_eps)

    def forward(self, x):
        x = self.patch_embed(x)
        x = self.encoder(x)
        x = self.norm(x)

        if self.patch_embed.cls_token is not None:
            feat = x[:, 0]
        else:
            feat = torch.mean(x, dim=1)
        return BackboneOutput(last_feature=feat)


def vit(task, conf_model_backbone):
    # ViT tiny
    return VisionTransformer(task, conf_model_backbone.params, conf_model_backbone.stage_params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import json
import os
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image
import torch
import torch.distributed as dist
from loguru import logger
from omegaconf import ListConfig

from .base import BaseCustomDataset, BaseSampleLoader
from .utils.constants import IMG_EXTENSIONS
from .utils.misc import get_detection_label, natural_key


class DetectionSampleLoader(BaseSampleLoader):
    def __init__(self, conf_data, train_valid_split_ratio):
        super(DetectionSampleLoader, self).__init__(conf_data, train_valid_split_ratio)

    def load_data(self, split='train'):
        assert split in ['train', 'valid', 'test'], f"split should be either {['train', 'valid', 'test']}."
        data_root = Path(self.conf_data.path.root)
        split_dir = self.conf_data.path[split]
        image_dir: Path = data_root / split_dir.image
        annotation_dir: Optional[Path] = data_root / split_dir.label if split_dir.label is not None else None
        images: List[str] = []
        labels: List[str] = []
        images_and_targets: List[Dict[str, str]] = []
        if annotation_dir is not None:
            for ext in IMG_EXTENSIONS:
                for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}')):
                    ann_path_maybe = annotation_dir / file.with_suffix('.txt').name
                    if not ann_path_maybe.exists():
                        continue
                    images.append(str(file))
                    labels.append(str(ann_path_maybe))
                # TODO: get paired data from regex pattern matching (self.conf_data.path.pattern)

            images = sorted(images, key=lambda k: natural_key(k))
            labels = sorted(labels, key=lambda k: natural_key(k))
            images_and_targets.extend([{'image': str(image), 'label': str(label)} for image, label in zip(images, labels)])

        else:
            for ext in IMG_EXTENSIONS:
                images_and_targets.extend([{'image': str(file), 'label': None}
                                        for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])
            images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k['image']))

        return images_and_targets

    def load_id_mapping(self):
        root_path = Path(self.conf_data.path.root)

        if isinstance(self.conf_data.id_mapping, ListConfig):
            return list(self.conf_data.id_mapping)

        elif isinstance(self.conf_data.id_mapping, str):
            id_mapping_path = root_path / self.conf_data.id_mapping
            if not os.path.isfile(id_mapping_path):
                raise FileNotFoundError(f"File not found: {id_mapping_path}")

            with open(id_mapping_path, 'r') as f:
                id_mapping = json.load(f)
            return id_mapping

        else:
            raise ValueError(f"Unsupported id_mapping value {self.conf_data.id_mapping}")

    def load_class_map(self, id_mapping):
        idx_to_class: Dict[int, str] = dict(enumerate(id_mapping))
        return {'idx_to_class': idx_to_class}

    def load_huggingface_samples(self):
        raise NotImplementedError


class DetectionCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, **kwargs):
        super(DetectionCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, **kwargs,
        )

    @staticmethod
    def xywhn2xyxy(original: np.ndarray, w: int, h: int, padw=0, padh=0):
        converted = original.copy()
        # left, top (lt)
        converted[..., 0] = w * (original[..., 0] - original[..., 2] / 2) + padw
        converted[..., 1] = h * (original[..., 1] - original[..., 3] / 2) + padh
        # right, bottom (rb)
        converted[..., 2] = w * (original[..., 0] + original[..., 2] / 2) + padw
        converted[..., 3] = h * (original[..., 1] + original[..., 3] / 2) + padh

        # bbox clamping
        np.clip(converted[..., 0::2], a_min=0, a_max=w, out=converted[..., 0::2])
        np.clip(converted[..., 1::2], a_min=0, a_max=h, out=converted[..., 1::2])
        return converted

    def cache_dataset(self, sampler, distributed):
        if (not distributed) or (distributed and dist.get_rank() == 0):
            logger.info(f'Caching | Loading samples of {self.mode} to memory... This can take minutes.')

        def _load(i, samples):
            image = Image.open(Path(samples[i]['image'])).convert('RGB')
            label = self.samples[i]['label']
            if label is not None:
                label = get_detection_label(Path(label))
            return i, image, label

        num_threads = 8 # TODO: Compute appropriate num_threads
        load_imgs = ThreadPool(num_threads).imap(
            partial(_load, samples=self.samples),
            sampler
        )
        for i, image, label in load_imgs:
            self.samples[i]['image'] = image
            self.samples[i]['label'] = label

        self.cache = True

    def __getitem__(self, index):
        if self.cache:
            img = self.samples[index]['image']
            ann = self.samples[index]['label']
        else:
            img = Image.open(self.samples[index]['image']).convert('RGB')
            ann_path = Path(self.samples[index]['label']) if self.samples[index]['label'] is not None else None
            ann = get_detection_label(Path(ann_path)) if ann_path is not None else None

        w, h = img.size

        outputs = {}
        outputs.update({'indices': index})
        if ann is None:
            out = self.transform(image=img)
            outputs.update({'pixel_values': out['image'], 'org_shape': (h, w)})
            return outputs

        label, boxes_yolo = ann
        boxes = self.xywhn2xyxy(boxes_yolo, w, h)

        out = self.transform(image=img, label=label, bbox=boxes, dataset=self)
        # Remove
        mask = np.minimum(out['bbox'][:, 2] - out['bbox'][:, 0], out['bbox'][:, 3] - out['bbox'][:, 1]) > 1
        out['bbox'] = out['bbox'][mask]
        out['label'] = torch.as_tensor(out['label'].ravel(), dtype=torch.int64)
        out['label'] = out['label'][mask]
        outputs.update({'pixel_values': out['image'], 'bbox': out['bbox'],
                        'label': out['label']})

        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        outputs.update({'org_shape': (h, w)})
        return outputs

    def pull_item(self, index):
        img_path = Path(self.samples[index]['image'])
        ann_path = Path(self.samples[index]['label']) if 'label' in self.samples[index] else None
        img = Image.open(str(img_path)).convert('RGB')

        org_img = img.copy()
        w, h = img.size
        if ann_path is None:
            return org_img, np.zeros(0, 1), np.zeros(0, 5)

        label, boxes_yolo = get_detection_label(Path(ann_path))
        boxes = self.xywhn2xyxy(boxes_yolo, w, h)

        return org_img, label, boxes



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import time
from typing import Optional, List, Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

from ...op.custom import BasicBlock, Bottleneck, ConvLayer
from ...op.pidnet import DAPPM, PAPPM, Bag, Light_Bag, PagFM, segmenthead
from ...utils import FXTensorType, PIDNetModelOutput

use_align_corners = False


class PIDNet(nn.Module):
    task = 'segmentation'
    def __init__(
        self, 
        params: Optional[Dict] = None
    ) -> None:
        super(PIDNet, self).__init__()
        self.name = params.nick_name
        
        num_classes = params.num_classes
        m = params.m
        n = params.n
        planes = params.channels
        ppm_planes = params.ppm_channels
        head_planes = params.head_channels
        is_training = params.is_training

        self.is_training = is_training

        # I Branch
        self.conv1 = nn.Sequential(
            ConvLayer(in_channels=3, out_channels=planes,
                      kernel_size=3, stride=2, padding=1,
                      norm_type='batch_norm', act_type='relu', bias=True),
            ConvLayer(in_channels=planes, out_channels=planes,
                      kernel_size=3, stride=2, padding=1,
                      norm_type='batch_norm', act_type='relu', bias=True)
        )

        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_layer(BasicBlock, planes, planes, m)
        self.layer2 = self._make_layer(BasicBlock, planes, planes * 2, m, stride=2)
        self.layer3 = self._make_layer(BasicBlock, planes * 2, planes * 4, n, stride=2)
        self.layer4 = self._make_layer(BasicBlock, planes * 4, planes * 8, n, stride=2)
        self.layer5 = self._make_layer(Bottleneck, planes * 8, planes * 8, 2, stride=2, expansion=2)

        # P Branch
        self.compression3 = ConvLayer(in_channels=planes * 4, out_channels=planes * 2,
                                      kernel_size=1, stride=1, padding=0,
                                      norm_type='batch_norm', use_act=False)

        self.compression4 = ConvLayer(in_channels=planes * 8, out_channels=planes * 2,
                                      kernel_size=1, stride=1, padding=0,
                                      norm_type='batch_norm', use_act=False)

        self.pag3 = PagFM(planes * 2, planes)
        self.pag4 = PagFM(planes * 2, planes)

        self.layer3_ = self._make_layer(BasicBlock, planes * 2, planes * 2, m)
        self.layer4_ = self._make_layer(BasicBlock, planes * 2, planes * 2, m)
        self.layer5_ = self._make_layer(Bottleneck, planes * 2, planes * 2, 1, expansion=2)

        # D Branch
        if m == 2:
            self.layer3_d = self._make_single_layer(BasicBlock, planes * 2, planes)
            self.layer4_d = self._make_layer(Bottleneck, planes, planes, 1, expansion=2)
            self.diff3 = ConvLayer(in_channels=planes * 4, out_channels=planes,
                                   kernel_size=3, stride=1, padding=1,
                                   norm_type='batch_norm', use_act=False)
            self.diff4 = ConvLayer(in_channels=planes * 8, out_channels=planes * 2,
                                   kernel_size=3, stride=1, padding=1,
                                   norm_type='batch_norm', use_act=False)
            self.spp = PAPPM(planes * 16, ppm_planes, planes * 4)
            self.dfm = Light_Bag(planes * 4, planes * 4)
        else:
            self.layer3_d = self._make_single_layer(BasicBlock, planes * 2, planes * 2)
            self.layer4_d = self._make_single_layer(BasicBlock, planes * 2, planes * 2)
            self.diff3 = ConvLayer(in_channels=planes * 4, out_channels=planes * 2,
                                   kernel_size=3, stride=1, padding=1,
                                   norm_type='batch_norm', use_act=False)
            self.diff4 = ConvLayer(in_channels=planes * 8, out_channels=planes * 2,
                                   kernel_size=3, stride=1, padding=1,
                                   norm_type='batch_norm', use_act=False)
            self.spp = DAPPM(planes * 16, ppm_planes, planes * 4)
            self.dfm = Bag(planes * 4, planes * 4)

        self.layer5_d = self._make_layer(Bottleneck, planes * 2, planes * 2, 1, expansion=2)

        # Prediction Head
        if self.is_training:
            self.seghead_p = segmenthead(planes * 2, head_planes, num_classes)
            self.seghead_d = segmenthead(planes * 2, planes, 1)

        self.final_layer = segmenthead(planes * 4, head_planes, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1, expansion=None):
        downsample = None
        if expansion is None:
            expansion = block.expansion
        if stride != 1 or inplanes != planes * expansion:
            downsample = ConvLayer(in_channels=inplanes, out_channels=planes * expansion,
                                   kernel_size=1, stride=stride,
                                   norm_type='batch_norm', use_act=False)

        layers = []
        layers.append(block(inplanes, planes, stride, downsample, expansion=expansion, no_out_act=True if block == Bottleneck else False))
        inplanes = planes * expansion
        for i in range(1, blocks):
            if i == (blocks-1):
                layers.append(block(inplanes, planes, stride=1, expansion=expansion, no_out_act=True))
            else:
                layers.append(block(inplanes, planes, stride=1, expansion=expansion, no_out_act=False))

        return nn.Sequential(*layers)

    def _make_single_layer(self, block, inplanes, planes, stride=1, expansion=None):
        downsample = None
        if expansion is None:
            expansion = block.expansion
        if stride != 1 or inplanes != planes * expansion:
            downsample = ConvLayer(in_channels=inplanes, out_channels=planes * expansion,
                                   kernel_size=1, stride=stride,
                                   norm_type='batch_norm', use_act=False)

        layer = block(inplanes, planes, stride, downsample, expansion=expansion, no_out_act=True)

        return layer

    @property
    def head_list(self):
        return ('seghead_p', 'seghead_d', 'final_layer')

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(self, x: FXTensorType, target=None):

        # assert H == x.size(2)
        # assert W == x.size(3)

        x = self.conv1(x)
        x = self.layer1(x)
        x = self.relu(self.layer2(self.relu(x)))
        x_ = self.layer3_(x)
        x_d = self.layer3_d(x)

        x = self.relu(self.layer3(x))
        x_ = self.pag3(x_, self.compression3(x))

        x_d = x_d + F.interpolate(
            self.diff3(x),
            size=x_d.shape[-2:],
            mode='bilinear', align_corners=use_align_corners)

        if not torch.jit.is_tracing() and not isinstance(x, torch.fx.Proxy):
            temp_p = x_

        x = self.relu(self.layer4(x))
        x_ = self.layer4_(self.relu(x_))
        x_d = self.layer4_d(self.relu(x_d))

        x_ = self.pag4(x_, self.compression4(x))

        x_d = x_d + F.interpolate(
            self.diff4(x),
            size=x_d.shape[-2:],
            mode='bilinear', align_corners=use_align_corners)

        if not torch.jit.is_tracing() and not isinstance(x, torch.fx.Proxy):
            temp_d = x_d

        x_ = self.layer5_(self.relu(x_))
        x_d = self.layer5_d(self.relu(x_d))

        x = F.interpolate(
            self.spp(self.layer5(x)),
            size=x_d.shape[-2:],
            mode='bilinear', align_corners=use_align_corners)

        x_ = self.final_layer(self.dfm(x_, x, x_d))

        if not torch.jit.is_tracing() and not isinstance(x, torch.fx.Proxy):
            x_extra_p = self.seghead_p(temp_p)
            x_extra_d = self.seghead_d(temp_d)
        else:
            x_extra_p = None
            x_extra_d = None

        return PIDNetModelOutput(extra_p=x_extra_p, extra_d=x_extra_d, pred=x_)

def pidnet(num_classes: int, conf_model_full) -> PIDNet:
    # PIDNet-S
    conf_model_full.num_classes = num_classes
    conf_model_full.is_training = True
    return PIDNet(params=conf_model_full)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path

from netspresso_trainer.evaluator_main import evaluation_cli
from netspresso_trainer.inferencer_main import inference_cli
from netspresso_trainer.trainer_main import parse_args_netspresso, train_cli, train_with_yaml

### Starting from v0.0.9, the default train function runs with yaml configuration
train = train_with_yaml
# train = train_with_config

version = (Path(__file__).parent / "VERSION").read_text().strip()

__version__ = version



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import math
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
from omegaconf import OmegaConf
from torch.utils.tensorboard import SummaryWriter
from torch.utils.tensorboard.summary import hparams


class TensorboardLogger:
    def __init__(self, task, model, result_dir, step_per_epoch: int, num_sample_images: int) -> None:
        super(TensorboardLogger, self).__init__()
        self.task = task
        self.model_name = model

        self.result_dir = Path(result_dir)
        self.result_dir.mkdir(exist_ok=True)
        self.step_per_epoch = step_per_epoch
        self.num_sample_images = num_sample_images

        self.tensorboard = SummaryWriter(self.result_dir / "tensorboard")

    def _as_numpy(self, value: Union[np.ndarray, torch.Tensor, list]) -> np.ndarray:
        if isinstance(value, np.ndarray):
            return value
        if isinstance(value, torch.Tensor):
            value = value.detach()
            value = value.cpu() if value.is_cuda else value
            value = value.numpy()
            return value
        if isinstance(value, list): # Pad images for tensorboard
            pad_shape = np.array([[v.shape[0], v.shape[1]] for v in value])
            pad_shape = pad_shape.max(0)
            ret_value = np.zeros((len(value), *pad_shape, 3), dtype=value[0].dtype)
            for i, v in enumerate(value):
                ret_value[i, :v.shape[0], :v.shape[1]] = v
            return ret_value

        raise TypeError(f"Unsupported type! {type(value)}")

    def _log_scalar(self, key: str, value, mode):
        global_step = self._epoch * self.step_per_epoch
        meta_string = f"{mode}/" if mode is not None else ""
        self.tensorboard.add_scalar(f"{meta_string}{key}", value, global_step=global_step)

    def _log_image(self, key: str, value: Union[np.ndarray, torch.Tensor], mode):
        global_step = self._epoch * self.step_per_epoch
        value = self._as_numpy(value)
        meta_string = f"{mode}/" if mode is not None else ""
        self.tensorboard.add_image(f"{meta_string}{key}", as_grid(value).astype(np.uint8),
                                   global_step=global_step, dataformats='HWC')

    def log_scalar(self, key, value, mode='train'):
        self._log_scalar(key, value, mode)

    def log_scalars_with_dict(self, scalar_dict, mode='train'):
        for k, v in scalar_dict.items():
            self._log_scalar(k, v, mode)

    def log_image(self, key, value: Union[np.ndarray, torch.Tensor], mode='train'):
        self._log_image(key, value, mode)

    def log_images_with_dict(self, image_dict, mode='train'):
        for k, v in image_dict.items():
            self._log_image(k, v, mode)

    def _get_rasterized_hparam(self, hparams):
        if not isinstance(hparams, dict):
            stem = hparams
            if not isinstance(hparams, (int, float, str, bool, torch.Tensor)):
                return str(stem)
            return stem

        rasterized_dict = {}
        for key, value in hparams.items():
            if isinstance(value, dict):
                rasterized_value = self._get_rasterized_hparam({f"{key}/{k}": v for k, v in value.items()})
                for k, v in rasterized_value.items():
                    rasterized_dict[k] = v
                continue
            rasterized_value = self._get_rasterized_hparam(value)
            rasterized_dict[key] = rasterized_value
        return rasterized_dict

    def log_hparams(self, hp_omegaconf: Union[Dict, List], final_metrics=None):

        if final_metrics is None:
            final_metrics = {}
        final_metrics = {f"hparams_metrics/{k}": v for k, v in final_metrics.items()}

        hp_dict = OmegaConf.to_container(hp_omegaconf, resolve=True)
        hp_for_log = self._get_rasterized_hparam(hp_dict)

        exp, ssi, sei = hparams(hparam_dict=hp_for_log, metric_dict=final_metrics)
        self.tensorboard.file_writer.add_summary(exp)
        self.tensorboard.file_writer.add_summary(ssi)
        self.tensorboard.file_writer.add_summary(sei)
        for k, v in final_metrics.items():
            self.tensorboard.add_scalar(k, v)

    def __call__(
        self,
        prefix: Literal['training', 'validation', 'evaluation', 'inference'],
        epoch: Optional[int] = None,
        images: Optional[List] = None,
        losses : Optional[Dict] = None,
        metrics: Optional[Dict] = None,
        learning_rate: Optional[float] = None,
        elapsed_time: Optional[float] = None,
        **kwargs
    ):
        self._epoch = 0 if epoch is None else epoch
        if losses is not None:
            self.log_scalars_with_dict(losses, mode=prefix)
        if metrics is not None:
            self.log_scalars_with_dict(metrics, mode=prefix)
        if isinstance(images, dict):  # TODO: array with multiple dicts
            self.log_images_with_dict(images, mode=prefix)

        if learning_rate is not None:
            self.log_scalar('learning_rate', learning_rate, mode='misc')
        if elapsed_time is not None:
            self.log_scalar('elapsed_time', elapsed_time, mode='misc')


def _as_grid(image_batch, rows: int, cols: int):
    _, h_image, w_image, channels = image_batch.shape
    canvas = np.zeros((h_image * rows, w_image * cols, channels))
    for idx, image in enumerate(image_batch):
        idx_row = idx // cols
        idx_col = idx % cols
        canvas[idx_row*h_image:(idx_row+1)*h_image,
               idx_col*w_image:(idx_col+1)*w_image] = image
    return canvas


def as_grid(image_batch, rows: Optional[int] = None, cols: Optional[int] = None):
    num_images, _, _, _ = image_batch.shape
    if rows is not None and cols is not None:
        assert num_images <= rows * cols
        return _as_grid(image_batch, rows, cols)
    if rows is not None:
        cols = math.ceil(num_images / rows)
        return _as_grid(image_batch, rows, cols)
    if cols is not None:
        rows = math.ceil(num_images / cols)
        return _as_grid(image_batch, rows, cols)

    square_size = math.ceil(math.sqrt(num_images))
    rows, cols = square_size, square_size
    return _as_grid(image_batch, rows, cols)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image
import torch
from omegaconf import OmegaConf

from ..utils.record import AverageMeter
from .image import ImageSaver
from .registry import VISUALIZER
from .stdout import StdOutLogger
from .tensorboard import TensorboardLogger
from .visualizer import magic_image_handler


class TrainingLogger():
    def __init__(
        self,
        conf,
        task: Literal['classification', 'segmentation', 'detection'],
        model: str,
        class_map: Dict[int, str],
        step_per_epoch: int,
        num_sample_images: int,
        result_dir: Union[Path, str],
        epoch: Optional[int] = None,
    ) -> None:
        super(TrainingLogger, self).__init__()
        self.conf = conf
        self.task: str = task
        self.model: str = model
        self.class_map: Dict = class_map
        self.epoch = epoch
        self.num_sample_images = num_sample_images

        self.project_id = conf.logging.project_id if conf.logging.project_id is not None else f"{self.task}_{self.model}"

        self._result_dir = result_dir
        OmegaConf.save(config=self.conf, f=(result_dir / "hparams.yaml"))

        self.use_tensorboard: bool = self.conf.logging.tensorboard
        self.use_imagesaver: bool = self.conf.logging.image
        self.use_stdout: bool = self.conf.logging.stdout

        self.loggers = []
        if self.use_imagesaver:
            self.loggers.append(ImageSaver(model=model, result_dir=self._result_dir))
        if self.use_tensorboard:
            self.tensorboard_logger = TensorboardLogger(task=task, model=model, result_dir=self._result_dir,
                                                        step_per_epoch=step_per_epoch, num_sample_images=num_sample_images)
            self.loggers.append(self.tensorboard_logger)
        if self.use_stdout:
            total_epochs = conf.training.epochs if hasattr(conf, 'training') else None
            self.loggers.append(StdOutLogger(task=task, model=model, total_epochs=total_epochs, result_dir=self._result_dir))

        if task in VISUALIZER:
            pallete = conf.data.pallete if 'pallete' in conf.data else None
            self.label_converter = VISUALIZER[task](class_map=class_map, pallete=pallete)

    @property
    def result_dir(self):
        return self._result_dir

    def update_epoch(self, epoch: int):
        self.epoch = epoch
        for logger in self.loggers:
            logger.epoch = self.epoch

    @staticmethod
    def _to_numpy(tensor: torch.Tensor):
        return tensor.detach().cpu().numpy()

    def _convert_scalar_as_readable(self, scalar_dict: Dict):
        for k, v in scalar_dict.items():
            if isinstance(v, (np.ndarray, float, int)):
                pass
                continue
            if isinstance(v, torch.Tensor):
                v_new = v.detach().cpu().numpy()
                scalar_dict.update({k: v_new})
                continue
            if isinstance(v, AverageMeter):
                v_new = v.avg
                scalar_dict.update({k: v_new})
                continue
            raise TypeError(f"Unsupported type for {k}!!! Current type: {type(v)}")
        return scalar_dict

    def _convert_imagedict_as_readable(self, images_dict: Dict):
        assert 'images' in images_dict
        image_new: np.ndarray = magic_image_handler(images_dict['images'])
        image_new = image_new.astype(np.uint8)
        images_dict.update({'images': image_new[:self.num_sample_images]})
        for k, v in images_dict.items():
            if k == 'images':
                continue

            # target, pred, bg_gt
            v = v[:self.num_sample_images]
            v_new: np.ndarray = magic_image_handler(
                self.label_converter(v, images=images_dict['images']))
            v_new = v_new.astype(np.uint8)
            images_dict.update({k: v_new})
        return images_dict

    def _convert_images_as_readable(self, images_dict_or_list: Union[Dict, List]):
        if isinstance(images_dict_or_list, list):
            images_list = images_dict_or_list

            if len(images_list) == 0:
                return None

            images_dict = {key: [] for key in images_list[0]}
            for minibatch in images_list:
                minibatch: Dict = self._convert_imagedict_as_readable(minibatch)
                for k_batch, v_batch in minibatch.items():
                    v_batch = list(v_batch)
                    if k_batch in images_dict:
                        images_dict[k_batch].extend(v_batch)
                        continue
                    images_dict[k_batch] = v_batch

            return images_dict

        if isinstance(images_dict_or_list, dict):
            images_dict = images_dict_or_list
            images_dict = self._convert_imagedict_as_readable(images_dict)
            return images_dict

        raise TypeError(f"Unsupported type for image logger!!! Current type: {type(images_dict_or_list)}")

    def log(
        self,
        prefix: Literal['training', 'validation', 'evaluation', 'inference'],
        epoch: Optional[int] = None,
        samples: Optional[List] = None,
        losses : Optional[Dict] = None,
        metrics: Optional[Dict] = None,
        learning_rate: Optional[float] = None,
        elapsed_time: Optional[float] = None,
    ):
        if not self.use_imagesaver: # TODO: This is uneffective way
            samples = None

        if losses is not None:
            losses = self._convert_scalar_as_readable(losses)
        if metrics is not None:
            metrics = self._convert_scalar_as_readable(metrics)
        if samples is not None:
            samples = self._convert_images_as_readable(samples)

        for logger in self.loggers:
            logger(
                prefix=prefix,
                epoch=epoch,
                losses=losses,
                metrics=metrics,
                images=samples,
                learning_rate=learning_rate,
                elapsed_time=elapsed_time
            )

    def log_end_of_traning(self, final_metrics=None):
        if final_metrics is None:
            final_metrics = {}
        if self.use_tensorboard:
            self.tensorboard_logger.log_hparams(self.conf, final_metrics=final_metrics)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Union

import torch
import torch.nn as nn
from torch import Tensor

from .environment import get_device

__all__ = ['save_onnx']


def _save_onnx(model: nn.Module, f: Union[str, Path], sample_input: Tensor,
               opset_version=13, input_names='images', output_names='output'):
    torch.onnx.export(model,  # model being run
                      sample_input,  # model input (or a tuple for multiple inputs)
                      f,  # where to save the model (can be a file or file-like object)
                      export_params=True,  # store the trained parameter weights inside the model file
                      opset_version=opset_version,  # the ONNX version to export the model to
                      do_constant_folding=True,  # whether to execute constant folding for optimization
                      input_names=[input_names],  # the model's input names
                      output_names=[output_names],  # the model's output names
                      dynamic_axes={input_names: {0: 'batch_size'},  # variable length axes
                                    output_names: {0: 'batch_size'}})


def save_onnx(model: nn.Module, f: Union[str, Path], sample_input: Tensor, opset_version):
    sample_input = sample_input.to(get_device(model))
    return _save_onnx(model, f, sample_input, opset_version=opset_version, input_names='images', output_names='output')



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

import numpy as np
import torch

from ..base import BaseMetric

TOPK_MAX = 20


@torch.no_grad()
def accuracy_topk(pred, target):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    maxk = pred.shape[-1]
    pred = pred.T
    class_num = pred.shape[0]
    correct = np.equal(pred, np.tile(target, (class_num, 1)))
    return lambda topk: correct[:min(topk, maxk)].reshape(-1).astype('float').sum(0)


class ClassificationMetric(BaseMetric):
    SUPPORT_METRICS: List[str] = ['Acc@1', 'Acc@5']

    def __init__(self, **kwargs):
        # TODO: Select metrics by user
        metric_names = ['Acc@1', 'Acc@5']
        primary_metric = 'Acc@1'

        assert set(metric_names).issubset(ClassificationMetric.SUPPORT_METRICS)
        super().__init__(metric_names=metric_names, primary_metric=primary_metric)

    def calibrate(self, pred, target, **kwargs):
        topk_callable = accuracy_topk(pred, target)

        Acc1_correct = topk_callable(topk=1)
        Acc5_correct = topk_callable(topk=5)

        self.metric_meter['Acc@1'].update(Acc1_correct, n=pred.shape[0])
        self.metric_meter['Acc@5'].update(Acc5_correct, n=pred.shape[0])



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import torch
import torch.nn as nn


# Copied from transformers.models.convnext.modeling_convnext.drop_path
def drop_path(input, drop_prob: float = 0.0, training: bool = False, scale_by_keep=True):
    """
    Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    Comment by Ross Wightman: This is the same as the DropConnect impl I created for EfficientNet, etc networks,
    however, the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for changing the
    layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use 'survival rate' as the
    argument.
    """
    if drop_prob == 0.0 or not training:
        return input
    keep_prob = 1 - drop_prob
    shape = (input.shape[0],) + (1,) * (input.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=input.dtype, device=input.device)
    random_tensor.floor_()  # binarize
    output = input.div(keep_prob) * random_tensor
    return output

class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob: float = 0., scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the YOLOX implementation of Megvii.
https://github.com/Megvii-BaseDetection/YOLOX/blob/main/yolox/models/yolo_head.py
"""
from typing import List
import math

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ....op.custom import ConvLayer, SeparableConvLayer
from ....utils import ModelOutput


class AnchorFreeDecoupledHead(nn.Module):
    def __init__(
        self,
        num_classes: int,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):
        super().__init__()
        act_type = params.act_type
        depthwise = params.depthwise

        self.num_classes = num_classes

        self.cls_convs = nn.ModuleList()
        self.reg_convs = nn.ModuleList()
        self.cls_preds = nn.ModuleList()
        self.reg_preds = nn.ModuleList()
        self.obj_preds = nn.ModuleList()
        self.stems = nn.ModuleList()
        Conv = SeparableConvLayer if depthwise else ConvLayer

        hidden_dim = int(intermediate_features_dim[0]) # 256 * width
        for i in range(len(intermediate_features_dim)):
            self.stems.append(
                ConvLayer(
                    in_channels=int(intermediate_features_dim[i]),
                    out_channels=hidden_dim,
                    kernel_size=1,
                    stride=1,
                    act_type=act_type,
                )
            )
            self.cls_convs.append(
                nn.Sequential(
                    *[
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                    ]
                )
            )
            self.reg_convs.append(
                nn.Sequential(
                    *[
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                        Conv(
                            in_channels=hidden_dim,
                            out_channels=hidden_dim,
                            kernel_size=3,
                            stride=1,
                            act_type=act_type,
                        ),
                    ]
                )
            )
            self.cls_preds.append(
                nn.Conv2d(
                    in_channels=hidden_dim,
                    out_channels=self.num_classes,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            )
            self.reg_preds.append(
                nn.Conv2d(
                    in_channels=hidden_dim,
                    out_channels=4,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            )
            self.obj_preds.append(
                nn.Conv2d(
                    in_channels=hidden_dim,
                    out_channels=1,
                    kernel_size=1,
                    stride=1,
                    padding=0,
                )
            )

        # Initialize
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03
        self.apply(init_bn)
        self.initialize_biases(1e-2)

    def initialize_biases(self, prior_prob):
        for conv in self.cls_preds:
            b = conv.bias.view(1, -1)
            b.data.fill_(-math.log((1 - prior_prob) / prior_prob))
            conv.bias = torch.nn.Parameter(b.view(-1), requires_grad=True)

        for conv in self.obj_preds:
            b = conv.bias.view(1, -1)
            b.data.fill_(-math.log((1 - prior_prob) / prior_prob))
            conv.bias = torch.nn.Parameter(b.view(-1), requires_grad=True)

    def forward(self, xin, targets=None):
        outputs = []

        for k, (cls_conv, reg_conv, x) in enumerate(zip(self.cls_convs, self.reg_convs, xin)):
            x = self.stems[k](x)
            cls_x = x
            reg_x = x

            cls_feat = cls_conv(cls_x)
            cls_output = self.cls_preds[k](cls_feat)

            reg_feat = reg_conv(reg_x)
            reg_output = self.reg_preds[k](reg_feat)
            obj_output = self.obj_preds[k](reg_feat)

            output = torch.cat([reg_output, obj_output, cls_output], 1)

            outputs.append(output)

        return ModelOutput(pred=outputs)


def anchor_free_decoupled_head(num_classes, intermediate_features_dim, conf_model_head, **kwargs) -> AnchorFreeDecoupledHead:
    return AnchorFreeDecoupledHead(num_classes=num_classes,
                                   intermediate_features_dim=intermediate_features_dim,
                                   params=conf_model_head.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Callable, Dict

from .custom.image_proc import (
    AutoAugment,
    CenterCrop,
    ColorJitter,
    HSVJitter,
    Normalize,
    Pad,
    PoseTopDownAffine,
    RandomCrop,
    RandomErasing,
    RandomHorizontalFlip,
    RandomIoUCrop,
    RandomResize,
    RandomResizedCrop,
    RandomVerticalFlip,
    RandomZoomOut,
    Resize,
    ToTensor,
    TrivialAugmentWide,
)
from .custom.mixing import Mixing
from .custom.mosaic import MosaicDetection

TRANSFORM_DICT: Dict[str, Callable] = {
    'centercrop': CenterCrop,
    'colorjitter': ColorJitter,
    'pad': Pad,
    'randomcrop': RandomCrop,
    'randomresizedcrop': RandomResizedCrop,
    'randomhorizontalflip': RandomHorizontalFlip,
    'randomresize': RandomResize,
    'randomverticalflip': RandomVerticalFlip,
    'randomerasing': RandomErasing,
    'randomioucrop': RandomIoUCrop,
    'randomzoomout': RandomZoomOut,
    'resize': Resize,
    'mixing': Mixing,
    'mosaicdetection': MosaicDetection,
    'trivialaugmentwide': TrivialAugmentWide,
    'autoaugment': AutoAugment,
    'hsvjitter': HSVJitter,
    'posetopdownaffine': PoseTopDownAffine,
    'totensor': ToTensor,
    'normalize': Normalize,
}



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Literal

import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig

from .registry import OPTIMIZER_DICT


def build_overwrited_dict(optimizer_conf):
    tmp_optimizer_conf = optimizer_conf.copy()
    del tmp_optimizer_conf['overwrite'] # Remove unnecessary field
    overwrited_config_dict = {
        'backbone': tmp_optimizer_conf.copy(),
        'neck': tmp_optimizer_conf.copy(),
        'head': tmp_optimizer_conf.copy(),
    }

    overwrite_config = optimizer_conf.overwrite
    for module_overwrite_config in overwrite_config:
        if not hasattr(module_overwrite_config, 'param_group'):
            raise ValueError(f"Overwrite config must have 'param_group' field: {module_overwrite_config}")

        param_group = module_overwrite_config.param_group

        if hasattr(module_overwrite_config, 'lr'):
            overwrited_config_dict[param_group].lr = module_overwrite_config.lr
        if hasattr(module_overwrite_config, 'weight_decay'):
            overwrited_config_dict[param_group].weight_decay = module_overwrite_config.weight_decay
        if hasattr(module_overwrite_config, 'no_bias_decay'):
            overwrited_config_dict[param_group].no_bias_decay = module_overwrite_config.no_bias_decay
        if hasattr(module_overwrite_config, 'no_norm_weight_decay'):
            overwrited_config_dict[param_group].no_norm_weight_decay = module_overwrite_config.no_norm_weight_decay

    return overwrited_config_dict


def separate_no_weights_decay(params: set, module_config: DictConfig):
    bias_params = set(filter(lambda s: 'bias' in s, params))
    norm_params = set(filter(lambda s: 'norm.weight' in s, params))
    no_decay_params = set()
    if module_config.no_bias_decay:
        no_decay_params |= bias_params
        params -= bias_params
    if module_config.no_norm_weight_decay:
        no_decay_params |= norm_params
        params -= norm_params
    return params, no_decay_params


def split_param_groups(model, overwrited_config_dict):
    backbone_params = set()
    neck_params = set()
    head_params = set()

    # Separate parameters by module
    named_params_dict = dict(model.named_parameters())
    for k in named_params_dict:
        if k.startswith('backbone'):
            backbone_params.add(k)
        elif k.startswith('neck'):
            neck_params.add(k)
        elif k.startswith('head'):
            head_params.add(k)
        else:
            raise ValueError(f"Unknown parameter group: {k}") # Only for NetsPresso Trainer defined model

    # Separate parameters by weight, bias, and norm
    backbone_params = separate_no_weights_decay(backbone_params, overwrited_config_dict['backbone'])
    neck_params = separate_no_weights_decay(neck_params, overwrited_config_dict['neck'])
    head_params = separate_no_weights_decay(head_params, overwrited_config_dict['head'])

    param_set_dict = {'backbone': backbone_params, 'neck': neck_params, 'head': head_params}

    # Build param_groups and param_configs
    param_groups = []
    param_opt_configs = []
    for param_group_key, module_overwrite_config in overwrited_config_dict.items():
        # Remove unnecessary fields
        del module_overwrite_config['no_bias_decay']
        del module_overwrite_config['no_norm_weight_decay']

        params, no_decay_params = param_set_dict[param_group_key]

        module_overwrite_config.group_name = param_group_key + '_group'
        if params: # Skip if no parameters, e.g. no neck in the model
            param_groups.append([named_params_dict[k] for k in params]) # Convert from key to parameters
            param_opt_configs.append(module_overwrite_config.copy())

        module_overwrite_config.group_name = param_group_key + '_no_decay_group'
        module_overwrite_config.weight_decay = 0.0
        if no_decay_params: # Skip if no parameters, e.g. use weight decay for all parameters
            param_groups.append([named_params_dict[k] for k in no_decay_params])
            param_opt_configs.append(module_overwrite_config.copy())

    return param_groups, param_opt_configs

def build_optimizer(
    model,
    single_task_model: bool,
    optimizer_conf: DictConfig,
):
    no_dist_model = model.module if hasattr(model, 'module') else model

    opt_name: Literal['sgd', 'adam', 'adamw', 'adamax', 'adadelta', 'adagrad', 'rmsprop'] = optimizer_conf.name.lower()
    assert opt_name in OPTIMIZER_DICT

    overwrite_config = optimizer_conf.overwrite
    if overwrite_config:
        assert not single_task_model, "Overwrite config only for non single task model"
        overwrited_config_dict = build_overwrited_dict(optimizer_conf)
        param_groups, param_opt_configs = split_param_groups(no_dist_model, overwrited_config_dict)

        base_param_group = param_groups.pop(0)
        base_opt_config = param_opt_configs.pop(0)
        optimizer = OPTIMIZER_DICT[opt_name](base_param_group, base_opt_config)
        optimizer.param_groups[0]['group_name'] = base_opt_config.group_name # Add group name manually

        for param_group, opt_config in zip(param_groups, param_opt_configs):
            del opt_config['name'] # Remove unnecessary field
            optimizer.add_param_group({'params': param_group, **opt_config})

    else:
        named_params_dict = dict(no_dist_model.named_parameters())

        params = set(named_params_dict.keys())
        params, no_decay_params = separate_no_weights_decay(params, optimizer_conf)

        params = [named_params_dict[k] for k in params]
        no_decay_params = [named_params_dict[k] for k in no_decay_params]

        optimizer = OPTIMIZER_DICT[opt_name](params, optimizer_conf)
        if no_decay_params:
            optimizer.add_param_group({'params': no_decay_params, 'weight_decay': 0.0})

    return optimizer



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
from functools import partial
from pathlib import Path
from typing import Dict, List, Literal, Optional, Type, Union

import torch.distributed as dist
import torch.utils.data as data
from loguru import logger
from omegaconf import DictConfig

from .augmentation.registry import TRANSFORM_DICT
from .registry import CREATE_TRANSFORM, CUSTOM_DATASET, DATA_SAMPLER, HUGGINGFACE_DATASET
from .utils.collate_fn import classification_mix_collate_fn, classification_onehot_collate_fn, detection_collate_fn
from .utils.loader import create_loader

TRAIN_VALID_SPLIT_RATIO = 0.9


def dataset_path_check(conf_data: DictConfig, mode: Literal['train', 'test']):
    if mode == 'train':
        train_check = (conf_data.path.train.image is not None) and (conf_data.path.train.label is not None)
        assert train_check, "For training, train split of dataset must be provided."

        if conf_data.path.test.image:
            logger.warning('For training, test split of dataset is not needed. This field will be ignored.')
        conf_data.path.test.image = None
        conf_data.path.test.label = None

    elif mode == 'test':
        assert conf_data.path.test.image is not None, "For test, test split of dataset must be provided."

        if conf_data.path.train.image:
            logger.warning('For test (evaluation or inference), train split of dataset is not needed. This field will be ignored.')
        conf_data.path.train.image = None
        conf_data.path.train.label = None

        if conf_data.path.valid.image:
            logger.warning('For test (evaluation or inference), valid split of dataset is not needed. This field will be ignored.')
        conf_data.path.valid.image = None
        conf_data.path.valid.label = None

    else:
        raise ValueError(f"mode of build_dataset cannot be {mode}. Must be one of ['train', 'test'].")


def loaded_dataset_check(
    conf_data: DictConfig,
    train_dataset: data.Dataset,
    valid_dataset: data.Dataset,
    test_dataset: data.Dataset,
    distributed: bool,
    mode: Literal['train', 'test']
):
    if mode == 'train':
        if not distributed or dist.get_rank() == 0:
            logger.info(f"Summary | Dataset: <{conf_data.name}> (with {conf_data.format} format)")
            logger.info(f"Summary | Training dataset: {len(train_dataset)} sample(s)")
            if valid_dataset is not None:
                logger.info(f"Summary | Validation dataset: {len(valid_dataset)} sample(s)")
        assert len(train_dataset) > 0, "Training dataset has no samples. Please check your dataset configuration."

    elif mode == 'test':
        if not distributed or dist.get_rank() == 0:
            logger.info(f"Summary | Dataset: <{conf_data.name}> (with {conf_data.format} format)")
            logger.info(f"Summary | Test dataset: {len(test_dataset)} sample(s)")
        assert len(test_dataset) > 0, "Test dataset has no samples. Please check your dataset configuration."

    else:
        raise ValueError(f"mode of build_dataset cannot be {mode}. Must be one of ['train', 'test'].")


def build_dataset(
    conf_data: DictConfig,
    conf_augmentation: DictConfig,
    task: str,
    model_name: str,
    distributed: bool,
    mode: Literal['train', 'test'],
):
    if not distributed or dist.get_rank() == 0:
        logger.info('-'*40)
        logger.info("Loading data...")

    task = conf_data.task

    assert task in DATA_SAMPLER, f"Data sampler for {task} is not yet supported!"

    train_transform = CREATE_TRANSFORM(model_name, is_training=True)
    target_transform = CREATE_TRANSFORM(model_name, is_training=False)

    data_format = conf_data.format

    assert data_format in ['local', 'huggingface'], f"No such data format named {data_format} in {['local', 'huggingface']}!"

    if data_format == 'local':
        dataset_path_check(conf_data=conf_data, mode=mode)

        assert task in CUSTOM_DATASET, f"Local dataset for {task} is not yet supported!"
        data_sampler = DATA_SAMPLER[task](conf_data, train_valid_split_ratio=TRAIN_VALID_SPLIT_RATIO)

        train_samples, valid_samples, test_samples, misc = data_sampler.load_samples()
        idx_to_class = misc['idx_to_class'] if 'idx_to_class' in misc else None
        label_value_to_idx = misc['label_value_to_idx'] if 'label_value_to_idx' in misc else None

        train_dataset = None
        if train_samples is not None:
            train_dataset = CUSTOM_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='train',
                samples=train_samples, transform=train_transform, label_value_to_idx=label_value_to_idx
            )

        valid_dataset = None
        if valid_samples is not None:
            valid_dataset = CUSTOM_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='valid',
                samples=valid_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

        test_dataset = None
        if test_samples is not None:
            test_dataset = CUSTOM_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='test',
                samples=test_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

    elif data_format == 'huggingface':
        assert task in CUSTOM_DATASET, f"HuggingFace dataset for {task} is not yet supported!"
        assert task in DATA_SAMPLER, f"Data sampler for {task} is not yet supported!"

        data_sampler = DATA_SAMPLER[task](conf_data, train_valid_split_ratio=TRAIN_VALID_SPLIT_RATIO)

        train_samples, valid_samples, test_samples, misc = data_sampler.load_huggingface_samples()
        idx_to_class = misc['idx_to_class'] if 'idx_to_class' in misc else None
        label_value_to_idx = misc['label_value_to_idx'] if 'label_value_to_idx' in misc else None

        # Assumed hugging face dataset always has training split
        train_dataset = HUGGINGFACE_DATASET[task](
            conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='train',
            huggingface_dataset=train_samples, transform=train_transform, label_value_to_idx=label_value_to_idx
        )

        valid_dataset = None
        if valid_samples is not None:
            valid_dataset = HUGGINGFACE_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='valid',
                huggingface_dataset=valid_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

        test_dataset = None
        if test_samples is not None:
            test_dataset = HUGGINGFACE_DATASET[task](
                conf_data, conf_augmentation, model_name, idx_to_class=idx_to_class, split='test',
                huggingface_dataset=test_samples, transform=target_transform, label_value_to_idx=label_value_to_idx
            )

    loaded_dataset_check(conf_data, train_dataset, valid_dataset, test_dataset, distributed, mode)
    return train_dataset, valid_dataset, test_dataset


def build_dataloader(conf, task: str, model_name: str, dataset, phase, profile=False):
    is_training = phase == 'train'

    #TODO: Temporarily set ``cache_data`` as optional since this is experimental
    cache_data = conf.environment.cache_data if hasattr(conf.environment, 'cache_data') else False

    if task == 'classification':
        # TODO: ``phase`` should be removed later.
        transforms = getattr(conf.augmentation, phase, None)
        if transforms:
            name = transforms[-1].name.lower()
            if name == 'mixing':
                mix_kwargs = list(transforms[-1].keys())
                mix_kwargs.remove('name')
                mix_kwargs = {k:transforms[-1][k] for k in mix_kwargs}
                mix_kwargs['num_classes'] = dataset.num_classes
                mix_transforms = TRANSFORM_DICT[name](**mix_kwargs)

                collate_fn = partial(classification_mix_collate_fn, mix_transforms=mix_transforms)
            else:
                collate_fn = partial(classification_onehot_collate_fn, num_classes=dataset.num_classes)
        else:
            collate_fn = partial(classification_onehot_collate_fn, num_classes=dataset.num_classes)

        dataloader = create_loader(
            dataset,
            conf.data.name,
            logger,
            batch_size=conf.environment.batch_size,
            is_training=is_training,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            cache_data=cache_data,
            kwargs=None
        )
    elif task == 'segmentation':
        collate_fn = None

        if phase == 'train':
            batch_size = conf.environment.batch_size
        else:
            batch_size = conf.environment.batch_size if model_name == 'pidnet' and not conf.distributed else 1

        dataloader = create_loader(
            dataset,
            conf.data.name,
            logger,
            batch_size=batch_size,
            is_training=is_training,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            cache_data=cache_data,
            kwargs=None
        )
    elif task == 'detection':
        collate_fn = detection_collate_fn

        if phase == 'train':
            batch_size = conf.environment.batch_size
        else:
            batch_size = conf.environment.batch_size if not conf.distributed else 2

        dataloader = create_loader(
            dataset,
            conf.data.name,
            logger,
            batch_size=batch_size,
            is_training=is_training,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            cache_data=cache_data,
            kwargs=None
        )
    elif task == 'pose_estimation':
        collate_fn = None

        dataloader = create_loader(
            dataset,
            conf.data.name,
            logger,
            batch_size=conf.environment.batch_size,
            is_training=is_training,
            num_workers=conf.environment.num_workers if not profile else 1,
            distributed=conf.distributed,
            collate_fn=collate_fn,
            pin_memory=False,
            world_size=conf.world_size,
            rank=conf.rank,
            cache_data=cache_data,
            kwargs=None
        )
    else:
        raise AssertionError(f"Task ({task}) is not understood!")

    return dataloader



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the RTMPose implementation of mmpose.
https://github.com/open-mmlab/mmpose
"""
from itertools import product
from typing import Dict, Optional

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torchvision.models.detection._utils import BoxCoder, Matcher
from torchvision.ops import boxes as box_ops

from ..common import SigmoidFocalLoss


class RTMCCLoss(nn.Module):
    def __init__(self, **kwargs) -> None:
        super().__init__()
        # TODO: Get from config
        self.beta = 10.
        self.label_softmax = True
        self.label_beta = 10.
        self.use_target_weight = True
        self.mask = None
        self.mask_weight = 1.
        self.simcc_split_ratio = 2.
        self.input_size = (256, 256)

        self.sigma = (5.66, 5.66)
        self.normalize = False

        self.log_softmax = nn.LogSoftmax(dim=1)
        self.kl_loss = nn.KLDivLoss(reduction='none')

    def _map_coordinates(
        self,
        keypoints,
        keypoints_visible,
    ):
        """Mapping keypoint coordinates into SimCC space."""

        keypoints_split = keypoints.clone()
        keypoints_split = torch.round(keypoints_split * self.simcc_split_ratio)
        keypoints_split = keypoints_split.to(torch.int64)
        keypoint_weights = keypoints_visible.clone()

        return keypoints_split, keypoint_weights

    def gaussian_smoothng(self, keypoints):
        keypoints_xy = keypoints[..., :2]
        keypoints_visible = keypoints[..., 2]

        N, K, _ = keypoints_xy.shape
        # TODO: Get from config
        w, h = self.input_size
        W = round(w * self.simcc_split_ratio)
        H = round(h * self.simcc_split_ratio)

        keypoints_split, keypoint_weights = self._map_coordinates(
            keypoints_xy, keypoints_visible)

        target_x = torch.zeros((N, K, W), dtype=torch.float32, device=keypoints.device)
        target_y = torch.zeros((N, K, H), dtype=torch.float32, device=keypoints.device)

        # 3-sigma rule
        radius = torch.tensor(self.sigma, device=keypoints.device) * 3

        # xy grid
        x = torch.arange(0, W, 1, dtype=torch.float32, device=keypoints.device)
        y = torch.arange(0, H, 1, dtype=torch.float32, device=keypoints.device)

        for n, k in product(range(N), range(K)):
            # skip unlabled keypoints
            if keypoints_visible[n, k] < 0.5:
                continue

            mu = keypoints_split[n, k]

            # check that the gaussian has in-bounds part
            left, top = mu - radius
            right, bottom = mu + radius + 1

            if left >= W or top >= H or right < 0 or bottom < 0:
                keypoint_weights[n, k] = 0
                continue

            mu_x, mu_y = mu

            target_x[n, k] = torch.exp(-((x - mu_x)**2) / (2 * self.sigma[0]**2))
            target_y[n, k] = torch.exp(-((y - mu_y)**2) / (2 * self.sigma[1]**2))

        if self.normalize:
            norm_value = self.sigma * torch.sqrt(torch.pi * 2)
            target_x /= norm_value[0]
            target_y /= norm_value[1]

        return target_x, target_y, keypoint_weights

    def criterion(self, dec_outs, labels):
        """Criterion function."""
        log_pt = self.log_softmax(dec_outs * self.beta)
        if self.label_softmax:
            labels = F.softmax(labels * self.label_beta, dim=1)
        loss = torch.mean(self.kl_loss(log_pt, labels), dim=1)
        return loss

    def forward(self, out: Dict, target: Dict) -> torch.Tensor:
        pred = out['pred']
        keypoints = target['keypoints']

        N, K, _ = pred.shape
        loss = 0

        coord_split = pred.shape[-1] // 2
        pred_x = pred[..., :coord_split]
        pred_y = pred[..., coord_split:]

        gt_x, gt_y, target_weights = self.gaussian_smoothng(keypoints)

        weight = target_weights.reshape(-1) if self.use_target_weight else 1.0

        for p, t in zip([pred_x, pred_y], [gt_x, gt_y]):
            p = p.reshape(-1, p.size(-1))
            t = t.reshape(-1, t.size(-1))

            t_loss = self.criterion(p, t).mul(weight)

            if self.mask is not None:
                t_loss = t_loss.reshape(N, K)
                t_loss[:, self.mask] = t_loss[:, self.mask] * self.mask_weight

            loss = loss + t_loss.sum()

        return loss / K



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

# Modified from YOLOPAFPN of YOLOX:
# https://github.com/Megvii-BaseDetection/YOLOX/blob/main/yolox/models/yolo_pafpn.py

from typing import List

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ...op.custom import ConvLayer, CSPLayer, SeparableConvLayer
from ...utils import BackboneOutput


class YOLOPAFPN(nn.Module):
    """
    YOLOv3 model. Darknet 53 is the default backbone of this model.
    """

    def __init__(
        self,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):
        super().__init__()
        depthwise = params.depthwise
        self.in_channels = intermediate_features_dim
        Conv = SeparableConvLayer if depthwise else ConvLayer

        depth = params.dep_mul
        act_type = params.act_type

        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        self.lateral_conv0 = ConvLayer(
            in_channels=int(self.in_channels[2]), 
            out_channels=int(self.in_channels[1]), 
            kernel_size=1, 
            stride=1,
            act_type=act_type
        )
        self.C3_p4 = CSPLayer(
            in_channels=int(2 * self.in_channels[1]),
            out_channels=int(self.in_channels[1]),
            n=round(3 * depth),
            shortcut=False,
            depthwise=depthwise,
            act_type=act_type,
        )  # cat

        self.reduce_conv1 = ConvLayer(
            in_channels=int(self.in_channels[1]), 
            out_channels=int(self.in_channels[0]), 
            kernel_size=1, 
            stride=1, 
            act_type=act_type
        )
        self.C3_p3 = CSPLayer(
            in_channels=int(2 * self.in_channels[0]),
            out_channels=int(self.in_channels[0]),
            n=round(3 * depth),
            shortcut=False,
            depthwise=depthwise,
            act_type=act_type,
        )

        # bottom-up conv
        self.bu_conv2 = Conv(
            in_channels=int(self.in_channels[0]), 
            out_channels=int(self.in_channels[0]), 
            kernel_size=3, 
            stride=2, 
            act_type=act_type
        )
        self.C3_n3 = CSPLayer(
            in_channels=int(2 * self.in_channels[0]),
            out_channels=int(self.in_channels[1]),
            n=round(3 * depth),
            shortcut=False,
            depthwise=depthwise,
            act_type=act_type,
        )

        # bottom-up conv
        self.bu_conv1 = Conv(
            in_channels=int(self.in_channels[1]), 
            out_channels=int(self.in_channels[1]), 
            kernel_size=3, 
            stride=2, 
            act_type=act_type
        )
        self.C3_n4 = CSPLayer(
            in_channels=int(2 * self.in_channels[1]),
            out_channels=int(self.in_channels[2]),
            n=round(3 * depth),
            shortcut=False,
            depthwise=depthwise,
            act_type=act_type,
        )

        self._intermediate_features_dim = self.in_channels

        # Initialize
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03
        self.apply(init_bn)

    def forward(self, inputs):
        """
        Args:
            inputs: input images.

        Returns:
            Tuple[Tensor]: FPN feature.
        """

        [x2, x1, x0] = inputs

        fpn_out0 = self.lateral_conv0(x0)  # 1024->512/32
        f_out0 = self.upsample(fpn_out0)  # 512/16
        f_out0 = torch.cat([f_out0, x1], 1)  # 512->1024/16
        f_out0 = self.C3_p4(f_out0)  # 1024->512/16

        fpn_out1 = self.reduce_conv1(f_out0)  # 512->256/16
        f_out1 = self.upsample(fpn_out1)  # 256/8
        f_out1 = torch.cat([f_out1, x2], 1)  # 256->512/8
        pan_out2 = self.C3_p3(f_out1)  # 512->256/8

        p_out1 = self.bu_conv2(pan_out2)  # 256->256/16
        p_out1 = torch.cat([p_out1, fpn_out1], 1)  # 256->512/16
        pan_out1 = self.C3_n3(p_out1)  # 512->512/16

        p_out0 = self.bu_conv1(pan_out1)  # 512->512/32
        p_out0 = torch.cat([p_out0, fpn_out0], 1)  # 512->1024/32
        pan_out0 = self.C3_n4(p_out0)  # 1024->1024/32

        outputs = (pan_out2, pan_out1, pan_out0)
        return BackboneOutput(intermediate_features=outputs)
    
    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

def yolopafpn(intermediate_features_dim, conf_model_neck, **kwargs):
    return YOLOPAFPN(intermediate_features_dim=intermediate_features_dim, params=conf_model_neck.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the RetinaNet implementation of torchvision.
https://github.com/pytorch/vision/blob/main/torchvision/models/detection/retinanet.py
"""
import math
from typing import List

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ....op.custom import ConvLayer
from ....utils import AnchorBasedDetectionModelOutput
from .detection import AnchorGenerator


class AnchorDecoupledHead(nn.Module):
    def __init__(
        self,
        num_classes: int,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):
        super().__init__()
        assert len(set(intermediate_features_dim)) == 1, "Feature dimensions of all stages have to same."
        in_channels = intermediate_features_dim[0]

        anchor_sizes = params.anchor_sizes
        aspect_ratios = params.aspect_ratios
        num_layers = params.num_layers

        norm_layer = params.norm_type

        aspect_ratios = (aspect_ratios,) * len(anchor_sizes)
        # TODO: Temporarily use hard-coded img_size
        self.anchor_generator = AnchorGenerator(anchor_sizes, aspect_ratios, (512, 512)) 
        num_anchors = self.anchor_generator.num_anchors_per_location()[0]

        self.classification_head = RetinaNetClassificationHead(
            in_channels, num_anchors, num_classes, num_layers, norm_layer=norm_layer
        )
        self.regression_head = RetinaNetRegressionHead(in_channels, num_anchors, num_layers, norm_layer=norm_layer)

    def forward(self, x):
        anchors = torch.cat(self.anchor_generator(x), dim=0)
        cls_logits = self.classification_head(x)
        bbox_regression = self.regression_head(x)
        return AnchorBasedDetectionModelOutput(anchors=anchors, cls_logits=cls_logits, bbox_regression=bbox_regression)


class RetinaNetClassificationHead(nn.Module):

    def __init__(
        self,
        in_channels,
        num_anchors,
        num_classes,
        num_layers,
        prior_probability = 0.01,
        norm_layer: str = 'batch_norm',
    ):
        super().__init__()

        conv = []
        for _ in range(num_layers):
            conv.append(
                ConvLayer(in_channels=in_channels, out_channels=in_channels,
                          kernel_size=3, stride=1, norm_type=norm_layer)
            )
        self.conv = nn.Sequential(*conv)

        for layer in self.conv.modules():
            if isinstance(layer, nn.Conv2d):
                torch.nn.init.normal_(layer.weight, std=0.01)
                if layer.bias is not None:
                    torch.nn.init.constant_(layer.bias, 0)

        self.cls_logits = nn.Conv2d(in_channels, num_anchors * num_classes, kernel_size=3, stride=1, padding=1)
        torch.nn.init.normal_(self.cls_logits.weight, std=0.01)
        torch.nn.init.constant_(self.cls_logits.bias, -math.log((1 - prior_probability) / prior_probability))

        self.num_classes = num_classes
        self.num_anchors = num_anchors

    def forward(self, x):
        all_cls_logits = []

        for features in x:
            cls_logits = self.conv(features)
            cls_logits = self.cls_logits(cls_logits)

            # Permute classification output from (N, A * K, H, W) to (N, HWA, K).
            N, _, H, W = cls_logits.shape
            cls_logits = cls_logits.view(N, -1, self.num_classes, H, W)
            cls_logits = cls_logits.permute(0, 3, 4, 1, 2)
            cls_logits = cls_logits.reshape(N, -1, self.num_classes)  # Size=(N, HWA, K)

            all_cls_logits.append(cls_logits)

        return all_cls_logits


class RetinaNetRegressionHead(nn.Module):

    def __init__(
        self, 
        in_channels,
        num_anchors,
        num_layers,
        norm_layer: str = 'batch_norm'
    ):
        super().__init__()

        conv = []
        for _ in range(num_layers):
            conv.append(
                ConvLayer(in_channels=in_channels, out_channels=in_channels,
                          kernel_size=3, stride=1, norm_type=norm_layer)
            )
        self.conv = nn.Sequential(*conv)

        self.bbox_reg = nn.Conv2d(in_channels, num_anchors * 4, kernel_size=3, stride=1, padding=1)
        torch.nn.init.normal_(self.bbox_reg.weight, std=0.01)
        torch.nn.init.zeros_(self.bbox_reg.bias)

        for layer in self.conv.modules():
            if isinstance(layer, nn.Conv2d):
                torch.nn.init.normal_(layer.weight, std=0.01)
                if layer.bias is not None:
                    torch.nn.init.zeros_(layer.bias)

    def forward(self, x, targets=None):
        all_bbox_regression = []

        for features in x:
            bbox_regression = self.conv(features)
            bbox_regression = self.bbox_reg(bbox_regression)

            # Permute bbox regression output from (N, 4 * A, H, W) to (N, HWA, 4).
            N, _, H, W = bbox_regression.shape
            bbox_regression = bbox_regression.view(N, -1, 4, H, W)
            bbox_regression = bbox_regression.permute(0, 3, 4, 1, 2)
            bbox_regression = bbox_regression.reshape(N, -1, 4)  # Size=(N, HWA, 4)

            all_bbox_regression.append(bbox_regression)

        return all_bbox_regression


def anchor_decoupled_head(num_classes, intermediate_features_dim, conf_model_head, **kwargs) -> AnchorDecoupledHead:
    return AnchorDecoupledHead(num_classes=num_classes,
                               intermediate_features_dim=intermediate_features_dim,
                               params=conf_model_head.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import torch.optim as optim


class Adadelta(optim.Adadelta):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        rho = optimizer_conf.rho
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, rho=rho, weight_decay=weight_decay)


class Adagrad(optim.Adagrad):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        lr_decay = optimizer_conf.lr_decay
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, lr_decay=lr_decay, weight_decay=weight_decay)


class Adam(optim.Adam):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        betas = optimizer_conf.betas
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, betas=betas, weight_decay=weight_decay)


class Adamax(optim.Adamax):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        betas = optimizer_conf.betas
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, betas=betas, weight_decay=weight_decay)


class AdamW(optim.AdamW):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        betas = optimizer_conf.betas
        weight_decay = optimizer_conf.weight_decay

        super().__init__(params=params, lr=lr, betas=betas, weight_decay=weight_decay)


class RMSprop(optim.RMSprop):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        alpha = optimizer_conf.alpha
        weight_decay = optimizer_conf.weight_decay
        momentum = optimizer_conf.momentum
        eps = optimizer_conf.eps

        super().__init__(params=params, lr=lr, alpha=alpha, weight_decay=weight_decay, momentum=momentum, eps=eps)


class SGD(optim.SGD):
    def __init__(
        self,
        params,
        optimizer_conf,
    ) -> None:
        lr = optimizer_conf.lr
        momentum = optimizer_conf.momentum
        weight_decay = optimizer_conf.weight_decay
        nesterov = optimizer_conf.nesterov

        super().__init__(params=params, lr=lr, momentum=momentum, weight_decay=weight_decay, nesterov=nesterov)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Literal, Optional

import torch
import torch.distributed as dist
from omegaconf import DictConfig

from .dataloaders import build_dataloader, build_dataset
from .models import SUPPORTING_TASK_LIST, build_model, is_single_task_model
from .pipelines import build_pipeline
from .utils.environment import set_device
from .utils.logger import add_file_handler, set_logger


def evaluation_common(
    conf: DictConfig,
    task: str,
    model_name: str,
    logging_dir: Path,
    log_level: Literal['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'] = 'INFO'
):
    distributed, world_size, rank, devices = set_device(conf.environment.seed)
    logger = set_logger(level=log_level, distributed=distributed)

    conf.distributed = distributed
    conf.world_size = world_size
    conf.rank = rank

    # Basic setup
    add_file_handler(logging_dir / "result.log", distributed=conf.distributed)

    if not distributed or dist.get_rank() == 0:
        logger.info(f"Task: {task} | Model: {model_name}")
        logger.info(f"Result will be saved at {logging_dir}")

    if conf.distributed and conf.rank != 0:
        torch.distributed.barrier()  # wait for rank 0 to download dataset

    single_task_model = is_single_task_model(conf.model)
    conf.model.single_task_model = single_task_model

    # Build dataloader
    _, _, test_dataset = build_dataset(conf.data, conf.augmentation, task, model_name, distributed=distributed, mode='test')

    if conf.distributed and conf.rank == 0:
        torch.distributed.barrier()

    eval_dataloader = build_dataloader(conf, task, model_name, dataset=test_dataset, phase='val')

    # Build model
    model = build_model(conf.model, test_dataset.num_classes, devices, conf.distributed)

    # Build evaluation pipeline
    pipeline_type = 'evaluation'
    pipeline = build_pipeline(pipeline_type=pipeline_type,
                              conf=conf,
                              task=task,
                              model_name=model_name,
                              model=model,
                              devices=devices,
                              class_map=test_dataset.class_map,
                              logging_dir=logging_dir,
                              is_graphmodule_training=None, # TODO: Remove is_graphmodule_training ...
                              dataloaders={'test': eval_dataloader})

    try:
        # Start evaluation
        pipeline.evaluation()

    except KeyboardInterrupt:
        pass
    except Exception as e:
        raise e



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import random
from functools import partial
from typing import Callable

import numpy as np
import torch
from torch.utils.data import DataLoader

from .constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from .misc import expand_to_chs
from .sampler import DistributedEvalSampler

NUM_RGB_CHANNEL = 3


def fast_collate(batch):
    """ A fast collation function optimized for uint8 images (np array or torch) and int64 targets (labels)"""
    assert isinstance(batch[0], tuple)
    batch_size = len(batch)
    if isinstance(batch[0][0], tuple):
        # This branch 'deinterleaves' and flattens tuples of input tensors into one tensor ordered by position
        # such that all tuple of position n will end up in a torch.split(tensor, batch_size) in nth position
        inner_tuple_size = len(batch[0][0])
        flattened_batch_size = batch_size * inner_tuple_size
        targets = torch.zeros(flattened_batch_size, dtype=torch.int64)
        tensor = torch.zeros((flattened_batch_size, *batch[0][0][0].shape), dtype=torch.uint8)
        for i in range(batch_size):
            assert len(batch[i][0]) == inner_tuple_size  # all input tensor tuples must be same length
            for j in range(inner_tuple_size):
                targets[i + j * batch_size] = batch[i][1]
                tensor[i + j * batch_size] += torch.from_numpy(batch[i][0][j])
        return tensor, targets
    elif isinstance(batch[0][0], np.ndarray):
        targets = torch.tensor([b[1] for b in batch], dtype=torch.int64)
        assert len(targets) == batch_size
        tensor = torch.zeros((batch_size, *batch[0][0].shape), dtype=torch.uint8)
        for i in range(batch_size):
            tensor[i] += torch.from_numpy(batch[i][0])
        return tensor, targets
    elif isinstance(batch[0][0], torch.Tensor):
        targets = torch.tensor([b[1] for b in batch], dtype=torch.int64)
        assert len(targets) == batch_size
        tensor = torch.zeros((batch_size, *batch[0][0].shape), dtype=torch.uint8)
        for i in range(batch_size):
            tensor[i].copy_(batch[i][0])
        return tensor, targets
    else:
        raise AssertionError()


def init_worker(worker_id, worker_seeding='all'):
    worker_info = torch.utils.data.get_worker_info()
    assert worker_info.id == worker_id
    if isinstance(worker_seeding, Callable):
        seed = worker_seeding(worker_info)
        random.seed(seed)
        torch.manual_seed(seed)
        np.random.seed(seed % (2 ** 32 - 1))
    else:
        assert worker_seeding in ('all', 'part')
        # random / torch seed already called in dataloader iter class w/ worker_info.seed
        # to reproduce some old results (same seed + hparam combo), partial seeding is required (skip numpy re-seed)
        if worker_seeding == 'all':
            np.random.seed(worker_info.seed % (2 ** 32 - 1))


def create_loader(
        dataset,
        dataset_name,
        logger,
        batch_size=1,
        is_training=False,
        num_aug_repeats=0,
        num_aug_splits=0,
        num_workers=1,
        distributed=False,
        collate_fn=None,
        pin_memory=False,
        fp16=False,
        tf_preprocessing=False,
        use_multi_epochs_loader=False,
        persistent_workers=True,
        worker_seeding='all',
        world_size=1,
        rank=0,
        cache_data=False,
        kwargs=None
):
    if is_training:
        sampler = torch.utils.data.distributed.DistributedSampler(dataset, num_replicas=world_size, rank=rank, drop_last=True)
    else:
        sampler = DistributedEvalSampler(dataset, num_replicas=world_size, rank=rank)

    if cache_data:
        dataset.cache_dataset(sampler, distributed)

    loader_args = {
        'batch_size': batch_size,
        'shuffle': not isinstance(dataset, torch.utils.data.IterableDataset) and sampler is None and is_training,
        'num_workers': num_workers,
        'sampler': sampler,
        'collate_fn': collate_fn,
        'pin_memory': pin_memory,
        'drop_last': is_training,
        'worker_init_fn': partial(init_worker, worker_seeding=worker_seeding),
        'persistent_workers': persistent_workers
    }
    try:
        loader = DataLoader(dataset, **loader_args)
    except TypeError:
        loader_args.pop('persistent_workers')  # only in Pytorch 1.7+
        loader = DataLoader(dataset, **loader_args)

    return loader



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------
"""
Based on the RT-DETR implementation of lyuwenyu.
https://github.com/lyuwenyu/RT-DETR/blob/main/rtdetr_pytorch/src/zoo/rtdetr/rtdetr_criterion.py
https://github.com/lyuwenyu/RT-DETR/blob/main/rtdetr_pytorch/src/zoo/rtdetr/matcher.py
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from scipy.optimize import linear_sum_assignment
from torchvision.ops.boxes import box_area


def box_iou(boxes1, boxes2):
    area1 = box_area(boxes1)
    area2 = box_area(boxes2)

    lt = torch.max(boxes1[:, None, :2], boxes2[:, :2])  # [N,M,2]
    rb = torch.min(boxes1[:, None, 2:], boxes2[:, 2:])  # [N,M,2]

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    inter = wh[:, :, 0] * wh[:, :, 1]  # [N,M]

    union = area1[:, None] + area2 - inter

    iou = inter / union
    return iou, union

def box_cxcywh_to_xyxy(x):
    x_c, y_c, w, h = x.unbind(-1)
    b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
         (x_c + 0.5 * w), (y_c + 0.5 * h)]
    return torch.stack(b, dim=-1)

def box_xyxy_to_cxcywh(x):
    x0, y0, x1, y1 = x.unbind(-1)
    b = [(x0 + x1) / 2, (y0 + y1) / 2,
         (x1 - x0), (y1 - y0)]
    return torch.stack(b, dim=-1)

def generalized_box_iou(boxes1, boxes2):
    """
    Generalized IoU from https://giou.stanford.edu/

    The boxes should be in [x0, y0, x1, y1] format

    Returns a [N, M] pairwise matrix, where N = len(boxes1)
    and M = len(boxes2)
    """
    # degenerate boxes gives inf / nan results
    # so do an early check
    assert (boxes1[:, 2:] >= boxes1[:, :2]).all()
    assert (boxes2[:, 2:] >= boxes2[:, :2]).all()
    iou, union = box_iou(boxes1, boxes2)

    lt = torch.min(boxes1[:, None, :2], boxes2[:, :2])
    rb = torch.max(boxes1[:, None, 2:], boxes2[:, 2:])

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    area = wh[:, :, 0] * wh[:, :, 1]

    return iou - (area - union) / area

"""
Based on the RT-DETR implementation of lyuwenyu.
"""
class HungarianMatcher(nn.Module):
    def __init__(self, weight_dict, alpha, gamma, use_focal_loss=False, **kwargs):
        super(HungarianMatcher, self).__init__()
        self.cost_class = weight_dict["cost_class"]
        self.cost_bbox = weight_dict["cost_bbox"]
        self.cost_giou = weight_dict["cost_giou"]

        self.use_focal_loss = use_focal_loss
        self.alpha = alpha
        self.gamma = gamma
        self.EPS = 1e-8

        assert self.cost_class != 0 or self.cost_bbox != 0 or self.cost_giou != 0, "All costs can't be zeros."

    @torch.no_grad()
    def forward(self, outputs, targets):
        bs, num_queries = outputs["pred_logits"].shape[:2]

        # We flatten to compute the cost matrices in a batch
        if self.use_focal_loss:
            out_prob = torch.sigmoid(outputs["pred_logits"].flatten(0, 1))
        else:
            out_prob = outputs["pred_logits"].flatten(0, 1).softmax(-1)  # [batch_size * num_queries, num_classes]

        out_bbox = outputs["pred_boxes"].flatten(0, 1)  # [batch_size * num_queries, 4]

        # Also concat the target labels and boxes
        tgt_ids = torch.cat([v["labels"] for v in targets])
        tgt_bbox = torch.cat([v["boxes"] for v in targets])

        # Compute the classification cost. Contrary to the loss, we don't use the NLL,
        # but approximate it in 1 - proba[target class].
        # The 1 is a constant that doesn't change the matching, it can be ommitted.
        if self.use_focal_loss:
            out_prob = out_prob[:, tgt_ids]
            neg_cost_class = (1 - self.alpha) * (out_prob**self.gamma) * (-(1 - out_prob + self.EPS).log())
            pos_cost_class = self.alpha * ((1 - out_prob)**self.gamma) * (-(out_prob + self.EPS).log())
            cost_class = pos_cost_class - neg_cost_class
        else:
            cost_class = -out_prob[:, tgt_ids]

        # Compute the L1 cost between boxes
        cost_bbox = torch.cdist(out_bbox, box_xyxy_to_cxcywh(tgt_bbox), p=1)

        # Compute the giou cost betwen boxes
        cost_giou = -generalized_box_iou(box_cxcywh_to_xyxy(out_bbox), tgt_bbox)

        # Final cost matrix
        C = self.cost_bbox * cost_bbox + self.cost_class * cost_class + self.cost_giou * cost_giou
        C = C.view(bs, num_queries, -1).cpu()

        sizes = [len(v["boxes"]) for v in targets]
        indices = [linear_sum_assignment(c[i]) for i, c in enumerate(C.split(sizes, -1))]

        return [(torch.as_tensor(i, dtype=torch.int64), torch.as_tensor(j, dtype=torch.int64)) for i, j in indices]


class DETRLoss(nn.Module):
    """ This class computes the loss for DETR.
    The process happens in two steps:
        1) we compute hungarian assignment between ground truth boxes and the outputs of the model
        2) we supervise each pair of matched ground-truth / prediction (supervise class and box)
    """

    def __init__(self, matcher, weight_dict, losses, alpha=0.2, gamma=2.0, cur_epoch=None, **kwargs):
        """ Create the criterion.
        Parameters:
            num_classes: number of object categories, omitting the special no-object category
            matcher: module able to compute a matching between targets and proposals
            weight_dict: dict containing as key the names of the losses and as values their relative weight.
            eos_coef: relative classification weight applied to the no-object category
            losses: list of all the losses to be applied. See get_loss for list of available losses.
        """
        super().__init__()
        self.matcher = eval(matcher.type)(matcher.weight_dict, matcher.alpha, matcher.gamma, matcher.use_focal_loss)
        self.weight_dict = weight_dict
        self.losses = losses

        self.alpha = alpha
        self.gamma = gamma

    def loss_labels(self, outputs, targets, indices, num_boxes, log=True):
        """Classification loss (NLL)
        targets dicts must contain the key "labels" containing a tensor of dim [nb_target_boxes]
        """
        assert 'pred_logits' in outputs
        src_logits = outputs['pred_logits']

        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.cat([t["labels"][J] for t, (_, J) in zip(targets, indices)])
        target_classes = torch.full(src_logits.shape[:2], self.num_classes,
                                    dtype=torch.int64, device=src_logits.device)
        target_classes[idx] = target_classes_o

        loss_ce = F.cross_entropy(src_logits.transpose(1, 2), target_classes, self.empty_weight)
        losses = {'loss_ce': loss_ce}

        if log:
            # TODO this should probably be a separate loss, not hacked in this one here
            losses['class_error'] = 100 - accuracy(src_logits[idx], target_classes_o)[0]
        return losses

    def loss_labels_focal(self, outputs, targets, indices, num_boxes, log=True):
        assert 'pred_logits' in outputs
        src_logits = outputs['pred_logits']

        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.cat([t["labels"][J] for t, (_, J) in zip(targets, indices)])
        target_classes = torch.full(src_logits.shape[:2], self.num_classes,
                                    dtype=torch.int64, device=src_logits.device)
        target_classes[idx] = target_classes_o

        target = F.one_hot(target_classes, num_classes=self.num_classes+1)[..., :-1]
        loss = torchvision.ops.sigmoid_focal_loss(src_logits, target, self.alpha, self.gamma, reduction='none')
        loss = loss.mean(1).sum() * src_logits.shape[1] / num_boxes

        return {'loss_focal': loss}

    def loss_labels_vfl(self, outputs, targets, indices, num_boxes, log=True):
        assert 'pred_boxes' in outputs
        idx = self._get_src_permutation_idx(indices)

        src_boxes = outputs['pred_boxes'][idx]
        target_boxes = torch.cat([t['boxes'][i] for t, (_, i) in zip(targets, indices)], dim=0)
        ious, _ = box_iou(box_cxcywh_to_xyxy(src_boxes), target_boxes)
        ious = torch.diag(ious).detach()

        src_logits = outputs['pred_logits']
        target_classes_o = torch.cat([t["labels"][J] for t, (_, J) in zip(targets, indices)])
        target_classes = torch.full(src_logits.shape[:2], self.num_classes,
                                    dtype=torch.int64, device=src_logits.device)
        target_classes[idx] = target_classes_o
        target = F.one_hot(target_classes, num_classes=self.num_classes + 1)[..., :-1]

        target_score_o = torch.zeros_like(target_classes, dtype=src_logits.dtype)
        target_score_o[idx] = ious.to(target_score_o.dtype)
        target_score = target_score_o.unsqueeze(-1) * target

        pred_score = torch.sigmoid(src_logits).detach()
        weight = self.alpha * pred_score.pow(self.gamma) * (1 - target) + target_score

        loss = F.binary_cross_entropy_with_logits(src_logits, target_score, weight=weight, reduction='none')
        loss = loss.mean(1).sum() * src_logits.shape[1] / num_boxes
        return {'loss_vfl': loss}

    @torch.no_grad()
    def loss_cardinality(self, outputs, targets, indices, num_boxes):
        """ Compute the cardinality error, ie the absolute error in the number of predicted non-empty boxes
        This is not really a loss, it is intended for logging purposes only. It doesn't propagate gradients
        """
        pred_logits = outputs['pred_logits']
        device = pred_logits.device
        tgt_lengths = torch.as_tensor([len(v["labels"]) for v in targets], device=device)
        # Count the number of predictions that are NOT "no-object" (which is the last class)
        card_pred = (pred_logits.argmax(-1) != pred_logits.shape[-1] - 1).sum(1)
        card_err = F.l1_loss(card_pred.float(), tgt_lengths.float())
        losses = {'cardinality_error': card_err}
        return losses

    def loss_boxes(self, outputs, targets, indices, num_boxes):
        """Compute the losses related to the bounding boxes, the L1 regression loss and the GIoU loss
           targets dicts must contain the key "boxes" containing a tensor of dim [nb_target_boxes, 4]
           The target boxes are expected in format (center_x, center_y, w, h), normalized by the image size.
        """
        assert 'pred_boxes' in outputs
        idx = self._get_src_permutation_idx(indices)
        src_boxes = outputs['pred_boxes'][idx]
        target_boxes = torch.cat([t['boxes'][i] for t, (_, i) in zip(targets, indices)], dim=0)
        losses = {}

        loss_bbox = F.l1_loss(src_boxes, box_xyxy_to_cxcywh(target_boxes), reduction='none')
        losses['loss_bbox'] = loss_bbox.sum() / num_boxes

        loss_giou = 1 - torch.diag(generalized_box_iou(
                box_cxcywh_to_xyxy(src_boxes),
                target_boxes))
        losses['loss_giou'] = loss_giou.sum() / num_boxes
        return losses

    def _get_src_permutation_idx(self, indices):
        # permute predictions following indices
        batch_idx = torch.cat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.cat([src for (src, _) in indices])
        return batch_idx, src_idx

    def _get_tgt_permutation_idx(self, indices):
        # permute targets following indices
        batch_idx = torch.cat([torch.full_like(tgt, i) for i, (_, tgt) in enumerate(indices)])
        tgt_idx = torch.cat([tgt for (_, tgt) in indices])
        return batch_idx, tgt_idx

    def get_loss(self, loss, outputs, targets, indices, num_boxes, **kwargs):
        loss_map = {
            'labels': self.loss_labels,
            'boxes': self.loss_boxes,
            'cardinality': self.loss_cardinality,
            'focal': self.loss_labels_focal,
            'vfl': self.loss_labels_vfl,
        }
        assert loss in loss_map, f'do you really want to compute {loss} loss?'
        return loss_map[loss](outputs, targets, indices, num_boxes, **kwargs)

    def forward(self, out, targets):
        """ This performs the loss computation.
        Parameters:
             outputs: dict of tensors, see the output specification of the model for the format
             targets: list of dicts, such that len(targets) == batch_size.
                      The expected keys in each dict depends on the losses applied, see each loss' doc
        """
        self.num_classes = targets['num_classes']
        img_size = targets['img_size']
        outputs = {'pred_boxes': out['pred'][..., :4], 'pred_logits': out['pred'][..., 4:], 'aux_outputs': out['aux_pred'], 'dn_aux_outputs': out['dn_aux_pred'], 'dn_meta': out['dn_meta']}
        empty_weight = torch.ones(self.num_classes + 1)
        empty_weight[-1] = 1e-4
        self.register_buffer('empty_weight', empty_weight)

        targets = targets['gt']
        # Normalize bounding boxes
        for idx, t in enumerate(targets):
            normalized_bboxes = self.normalize_bboxes(t['boxes'], img_size)
            t['boxes'] = normalized_bboxes
            targets[idx] = t
        outputs_without_aux = {k: v for k, v in outputs.items() if 'aux' not in k}

        # Retrieve the matching between the outputs of the last layer and the targets
        indices = self.matcher(outputs_without_aux, targets)

        # Compute the average number of target boxes accross all nodes, for normalization purposes
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor([num_boxes], dtype=torch.float, device=next(iter(outputs.values())).device)
        if torch.distributed.is_available() and torch.distributed.is_initialized():
            world_size = torch.distributed.get_world_size()
            torch.distributed.all_reduce(num_boxes)
        else:
            world_size = 1
        num_boxes = torch.clamp(num_boxes / world_size, min=1).item()

        # Compute all the requested losses
        losses = {}
        for loss in self.losses:
            l_dict = self.get_loss(loss, outputs, targets, indices, num_boxes)
            l_dict = {k: l_dict[k] * self.weight_dict[k] for k in l_dict if k in self.weight_dict}
            losses.update(l_dict)

        # In case of auxiliary losses, we repeat this process with the output of each intermediate layer.
        if outputs['aux_outputs'] is not None:
            for i, aux_outputs in enumerate(outputs['aux_outputs']):
                indices = self.matcher(aux_outputs, targets)
                for loss in self.losses:
                    if loss == 'masks':
                        # Intermediate masks losses are too costly to compute, we ignore them.
                        continue
                    kwargs = {}
                    if loss == 'labels':
                        # Logging is enabled only for the last layer
                        kwargs = {'log': False}

                    l_dict = self.get_loss(loss, aux_outputs, targets, indices, num_boxes, **kwargs)
                    l_dict = {k: l_dict[k] * self.weight_dict[k] for k in l_dict if k in self.weight_dict}
                    l_dict = {k + f'_aux_{i}': v for k, v in l_dict.items()}
                    losses.update(l_dict)

        # In case of cdn auxiliary losses. For rtdetr
        if outputs['dn_aux_outputs'] is not None:
            assert 'dn_meta' in outputs, ''
            indices = self.get_cdn_matched_indices(outputs['dn_meta'], targets)
            num_boxes = num_boxes * outputs['dn_meta']['dn_num_group']

            for i, aux_outputs in enumerate(outputs['dn_aux_outputs']):
                for loss in self.losses:
                    if loss == 'masks':
                        # Intermediate masks losses are too costly to compute, we ignore them.
                        continue
                    kwargs = {}
                    if loss == 'labels':
                        # Logging is enabled only for the last layer
                        kwargs = {'log': False}

                    l_dict = self.get_loss(loss, aux_outputs, targets, indices, num_boxes, **kwargs)
                    l_dict = {k: l_dict[k] * self.weight_dict[k] for k in l_dict if k in self.weight_dict}
                    l_dict = {k + f'_dn_{i}': v for k, v in l_dict.items()}
                    losses.update(l_dict)

        total_loss = torch.stack(list(losses.values())).sum()
        return total_loss

    @staticmethod
    def get_cdn_matched_indices(dn_meta, targets):
        '''get_cdn_matched_indices
        '''
        dn_positive_idx, dn_num_group = dn_meta["dn_positive_idx"], dn_meta["dn_num_group"]
        num_gts = [len(t['labels']) for t in targets]
        device = targets[0]['labels'].device

        dn_match_indices = []
        for i, num_gt in enumerate(num_gts):
            if num_gt > 0:
                gt_idx = torch.arange(num_gt, dtype=torch.int64, device=device)
                gt_idx = gt_idx.tile(dn_num_group)
                assert len(dn_positive_idx[i]) == len(gt_idx)
                dn_match_indices.append((dn_positive_idx[i], gt_idx))
            else:
                dn_match_indices.append((torch.zeros(0, dtype=torch.int64, device=device), \
                    torch.zeros(0, dtype=torch.int64,  device=device)))

        return dn_match_indices

    def normalize_bboxes(self, bboxes, img_size):
        """ Normalize bounding boxes to [0, 1] range """
        if isinstance(img_size, int):
            h = w = img_size
        else:
            h, w = img_size
        scale = torch.tensor([w, h, w, h], dtype=bboxes.dtype, device=bboxes.device)
        return bboxes / scale.unsqueeze(0)


@torch.no_grad()
def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    if target.numel() == 0:
        return [torch.zeros([], device=output.device)]
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import os
import random
from pathlib import Path
from typing import Union

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.nn as nn

__all__ = ['set_device', 'get_device']


def set_device(seed):
    # Torch settings
    cudnn.enabled = False

    # Reproducibility (except cudnn deterministicity)
    if seed is not None:
        torch.manual_seed(seed)  # for cpu
        torch.cuda.manual_seed(seed)  # for gpu
        random.seed(seed)
        np.random.seed(seed)

    # prepare device
    devices = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # distributed
    distributed = 'WORLD_SIZE' in os.environ and int(os.environ['WORLD_SIZE']) > 1
    world_size = 1
    rank = 0  # global rank
    if distributed:
        assert seed is not None and cudnn.benchmark is False, "distributed training requires reproducibility"
        dist.init_process_group(backend='nccl', init_method='env://')
        rank = dist.get_rank()
        devices = torch.device(f'cuda:{rank}')
        torch.cuda.set_device(rank)
        world_size = dist.get_world_size()
    assert rank >= 0

    return distributed, world_size, rank, devices


def get_device(x: Union[torch.Tensor, nn.Module]):
    if isinstance(x, torch.Tensor):
        return x.device
    if isinstance(x, nn.Module):
        return next(x.parameters()).device
    raise RuntimeError(f'{type(x)} do not have `device`')



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the RTMPose implementation of mmpose.
https://github.com/open-mmlab/mmpose
"""
import math
from typing import List, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from omegaconf import DictConfig

from ....op.custom import ConvLayer
from ....op.depth import DropPath
from ....op.registry import ACTIVATION_REGISTRY
from ....utils import FXTensorListType, ModelOutput


def rope(x, dim):
    """Applies Rotary Position Embedding to input tensor.

    Args:
        x (torch.Tensor): Input tensor.
        dim (int | list[int]): The spatial dimension(s) to apply
            rotary position embedding.

    Returns:
        torch.Tensor: The tensor after applying rotary position
            embedding.

    Reference:
        `RoFormer: Enhanced Transformer with Rotary
        Position Embedding <https://arxiv.org/abs/2104.09864>`_
    """
    shape = x.shape
    if isinstance(dim, int):
        dim = [dim]

    spatial_shape = [shape[i] for i in dim]
    total_len = 1
    for i in spatial_shape:
        total_len *= i

    position = torch.reshape(
        torch.arange(total_len, dtype=torch.int, device=x.device),
        spatial_shape)

    for _ in range(dim[-1] + 1, len(shape) - 1, 1):
        position = torch.unsqueeze(position, dim=-1)

    half_size = shape[-1] // 2
    freq_seq = -torch.arange(
        half_size, dtype=torch.int, device=x.device) / float(half_size)
    inv_freq = 10000**-freq_seq

    sinusoid = position[..., None] * inv_freq[None, None, :]

    sin = torch.sin(sinusoid)
    cos = torch.cos(sinusoid)
    x1, x2 = torch.chunk(x, 2, dim=-1)

    return torch.cat([x1 * cos - x2 * sin, x2 * cos + x1 * sin], dim=-1)


class Scale(nn.Module):
    """Scale vector by element multiplications.

    Args:
        dim (int): The dimension of the scale vector.
        init_value (float, optional): The initial value of the scale vector.
            Defaults to 1.0.
        trainable (bool, optional): Whether the scale vector is trainable.
            Defaults to True.
    """

    def __init__(self, dim, init_value=1., trainable=True):
        super().__init__()
        self.scale = nn.Parameter(
            init_value * torch.ones(dim), requires_grad=trainable)

    def forward(self, x):
        """Forward function."""

        return x * self.scale


class ScaleNorm(nn.Module):
    """Scale Norm.

    Args:
        dim (int): The dimension of the scale vector.
        eps (float, optional): The minimum value in clamp. Defaults to 1e-5.

    Reference:
        `Transformers without Tears: Improving the Normalization
        of Self-Attention <https://arxiv.org/abs/1910.05895>`_
    """

    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.scale = dim**-0.5
        self.eps = eps
        self.g = nn.Parameter(torch.ones(1))

    def forward(self, x):
        """Forward function.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: The tensor after applying scale norm.
        """
        norm = torch.norm(x, dim=-1, keepdim=True)
        norm = norm * self.scale
        return x / norm.clamp(min=self.eps) * self.g


# TODO: Replace RTMCCBlock as universal class of Gated Attention Unit
class RTMCCBlock(nn.Module):
    """Gated Attention Unit (GAU) in RTMBlock.

    Args:
        num_token (int): The number of tokens.
        in_token_dims (int): The input token dimension.
        out_token_dims (int): The output token dimension.
        expansion_factor (int, optional): The expansion factor of the
            intermediate token dimension. Defaults to 2.
        s (int, optional): The self-attention feature dimension.
            Defaults to 128.
        eps (float, optional): The minimum value in clamp. Defaults to 1e-5.
        dropout_rate (float, optional): The dropout rate. Defaults to 0.0.
        drop_path (float, optional): The drop path rate. Defaults to 0.0.
        attn_type (str, optional): Type of attention which should be one of
            the following options:

            - 'self-attn': Self-attention.
            - 'cross-attn': Cross-attention.

            Defaults to 'self-attn'.
        act_fn (str, optional): The activation function which should be one
            of the following options:

            - 'ReLU': ReLU activation.
            - 'SiLU': SiLU activation.

            Defaults to 'SiLU'.
        bias (bool, optional): Whether to use bias in linear layers.
            Defaults to False.
        use_rel_bias (bool, optional): Whether to use relative bias.
            Defaults to True.
        pos_enc (bool, optional): Whether to use rotary position
            embedding. Defaults to False.

    Reference:
        `Transformer Quality in Linear Time
        <https://arxiv.org/abs/2202.10447>`_
    """

    def __init__(self,
                 num_token,
                 in_token_dims,
                 out_token_dims,
                 expansion_factor=2,
                 s=128,
                 eps=1e-5,
                 dropout_rate=0.,
                 drop_path=0.,
                 attn_type='self-attn',
                 act_type='silu',
                 bias=False,
                 use_rel_bias=True,
                 pos_enc=False):

        super(RTMCCBlock, self).__init__()
        self.s = s
        self.num_token = num_token
        self.use_rel_bias = use_rel_bias
        self.attn_type = attn_type
        self.pos_enc = pos_enc
        self.drop_path = DropPath(drop_path) \
            if drop_path > 0. else nn.Identity()

        self.e = int(in_token_dims * expansion_factor)
        if use_rel_bias:
            if attn_type == 'self-attn':
                self.w = nn.Parameter(
                    torch.rand([2 * num_token - 1], dtype=torch.float))
            else:
                self.a = nn.Parameter(torch.rand([1, s], dtype=torch.float))
                self.b = nn.Parameter(torch.rand([1, s], dtype=torch.float))
        self.o = nn.Linear(self.e, out_token_dims, bias=bias)

        if attn_type == 'self-attn':
            self.uv = nn.Linear(in_token_dims, 2 * self.e + self.s, bias=bias)
            self.gamma = nn.Parameter(torch.rand((2, self.s)))
            self.beta = nn.Parameter(torch.rand((2, self.s)))
        else:
            self.uv = nn.Linear(in_token_dims, self.e + self.s, bias=bias)
            self.k_fc = nn.Linear(in_token_dims, self.s, bias=bias)
            self.v_fc = nn.Linear(in_token_dims, self.e, bias=bias)
            nn.init.xavier_uniform_(self.k_fc.weight)
            nn.init.xavier_uniform_(self.v_fc.weight)

        self.ln = ScaleNorm(in_token_dims, eps=eps)

        nn.init.xavier_uniform_(self.uv.weight)

        if act_type == 'silu' or act_type == 'relu':
            self.act_fn = ACTIVATION_REGISTRY[act_type]()
        else:
            raise NotImplementedError

        if in_token_dims == out_token_dims:
            self.shortcut = True
            self.res_scale = Scale(in_token_dims)
        else:
            self.shortcut = False

        self.sqrt_s = math.sqrt(s)

        self.dropout_rate = dropout_rate

        if dropout_rate > 0.:
            self.dropout = nn.Dropout(dropout_rate)

    def rel_pos_bias(self, seq_len, k_len=None):
        """Add relative position bias."""

        if self.attn_type == 'self-attn':
            t = F.pad(self.w[:2 * seq_len - 1], [0, seq_len]).repeat(seq_len)
            t = t[..., :-seq_len].reshape(-1, seq_len, 3 * seq_len - 2)
            r = (2 * seq_len - 1) // 2
            t = t[..., r:-r]
        else:
            a = rope(self.a.repeat(seq_len, 1), dim=0)
            b = rope(self.b.repeat(k_len, 1), dim=0)
            t = torch.bmm(a, b.permute(0, 2, 1))
        return t

    def _forward(self, inputs):
        """GAU Forward function."""

        if self.attn_type == 'self-attn':
            x = inputs
        else:
            x, k, v = inputs

        x = self.ln(x)

        # [B, K, in_token_dims] -> [B, K, e + e + s]
        uv = self.uv(x)
        uv = self.act_fn(uv)

        if self.attn_type == 'self-attn':
            # [B, K, e + e + s] -> [B, K, e], [B, K, e], [B, K, s]
            u, v, base = torch.split(uv, [self.e, self.e, self.s], dim=2)
            # [B, K, 1, s] * [1, 1, 2, s] + [2, s] -> [B, K, 2, s]
            base = base.unsqueeze(2) * self.gamma[None, None, :] + self.beta

            if self.pos_enc:
                base = rope(base, dim=1)
            # [B, K, 2, s] -> [B, K, s], [B, K, s]
            q, k = torch.unbind(base, dim=2)

        else:
            # [B, K, e + s] -> [B, K, e], [B, K, s]
            u, q = torch.split(uv, [self.e, self.s], dim=2)

            k = self.k_fc(k)  # -> [B, K, s]
            v = self.v_fc(v)  # -> [B, K, e]

            if self.pos_enc:
                q = rope(q, 1)
                k = rope(k, 1)

        # [B, K, s].permute() -> [B, s, K]
        # [B, K, s] x [B, s, K] -> [B, K, K]
        qk = torch.bmm(q, k.permute(0, 2, 1))

        if self.use_rel_bias:
            if self.attn_type == 'self-attn':
                bias = self.rel_pos_bias(q.size(1))
            else:
                bias = self.rel_pos_bias(q.size(1), k.size(1))
            qk += bias[:, :q.size(1), :k.size(1)]
        # [B, K, K]
        kernel = torch.square(F.relu(qk / self.sqrt_s))

        if self.dropout_rate > 0.:
            kernel = self.dropout(kernel)
        # [B, K, K] x [B, K, e] -> [B, K, e]
        x = u * torch.bmm(kernel, v)
        # [B, K, e] -> [B, K, out_token_dims]
        x = self.o(x)

        return x

    def forward(self, x):
        """Forward function."""

        if self.shortcut:
            res_shortcut = x[0] if self.attn_type == 'cross-attn' else x
            main_branch = self.drop_path(self._forward(x))
            return self.res_scale(res_shortcut) + main_branch
        else:
            return self.drop_path(self._forward(x))


class RTMCC(nn.Module):
    def __init__(
            self,
            num_classes: int,
            intermediate_features_dim: List[int],
            params: DictConfig,
    ):
        super().__init__()

        conv_kernel = params.conv_kernel
        attention_channels = params.attention_channels
        attnetion_act_type = params.attention_act_type
        attention_pos_enc = params.attention_pos_enc
        s = params.s
        expansion_factor = params.expansion_factor
        dropout_rate = params.dropout_rate
        drop_path = params.drop_path
        use_rel_bias = params.use_rel_bias

        self.simcc_split_ratio = params.simcc_split_ratio
        self.target_size = params.target_size

        # TODO: Get from backbone info
        flatten_dims = (self.target_size[0] // params.backbone_stride) * (self.target_size[1] // params.backbone_stride)

        # Define SimCC layers
        self.final_layer = nn.Conv2d(
            intermediate_features_dim[-1],
            num_classes,
            kernel_size=conv_kernel,
            stride=1,
            padding=conv_kernel // 2)
        self.mlp = nn.Sequential(
            ScaleNorm(flatten_dims),
            nn.Linear(flatten_dims, attention_channels, bias=False))

        W = int(self.target_size[1] * self.simcc_split_ratio)
        H = int(self.target_size[0] * self.simcc_split_ratio)

        self.gau = RTMCCBlock(
            num_classes,
            attention_channels,
            attention_channels,
            s=s,
            expansion_factor=expansion_factor,
            dropout_rate=dropout_rate,
            drop_path=drop_path,
            attn_type='self-attn',
            act_type=attnetion_act_type,
            use_rel_bias=use_rel_bias,
            pos_enc=attention_pos_enc)

        self.cls_x = nn.Linear(attention_channels, W, bias=False)
        self.cls_y = nn.Linear(attention_channels, H, bias=False)

    def forward(self, encoder_hidden_states: FXTensorListType, targets=None):
        out = encoder_hidden_states[-1]

        out = self.final_layer(out)  # -> B, K, H, W

        # flatten the output heatmap
        out = torch.flatten(out, 2)

        out = self.mlp(out)  # -> B, K, hidden

        out = self.gau(out)

        out_x = self.cls_x(out)
        out_y = self.cls_y(out)

        out = torch.cat([out_x, out_y], dim=-1)
        return ModelOutput(pred=out)


def rtmcc(num_classes, intermediate_features_dim, conf_model_head, **kwargs) -> RTMCC:
    return RTMCC(num_classes,
                 intermediate_features_dim,
                 params=conf_model_head.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List, Dict, Optional

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor

from ...op.custom import ConvLayer, UniversalInvertedResidualBlock
from ...op.base_metaformer import MobileMultiQueryAttention2D
from ...utils import BackboneOutput
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['mobilenetv4']

SUPPORTING_TASK = ['classification', 'segmentation', 'detection', 'pose_estimation']


class FusedIB(nn.Module):
    # Based on MobileNetV4: https://arxiv.org/pdf/2404.10518
    # Only for MobileNetV4
    def __init__(self, in_channel, hidden_channel, out_channel, kernel_size, stride, norm_type, act_type):
        super().__init__()
        self.block = []
        self.block.append(ConvLayer(in_channel, hidden_channel, kernel_size=kernel_size, 
                                    stride=stride, bias=False, norm_type=norm_type, act_type=act_type))
        self.block.append(ConvLayer(hidden_channel, out_channel, kernel_size=1, 
                                    stride=1, bias=False, norm_type=norm_type, use_act=False))
        self.block = nn.Sequential(*self.block)

    def forward(self, x):
        return self.block(x)


class MobileNetV4(nn.Module):

    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'MobileNetV4 is not supported on {self.task} task now.'
        self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        super(MobileNetV4, self).__init__()

        # Parameters
        stem_out_channel = params.stem_out_channel
        stem_kernel_size = params.stem_kernel_size
        stem_stride = params.stem_stride
        final_conv_out_channel = params.final_conv_out_channel
        final_conv_kernel_size = params.final_conv_kernel_size
        final_conv_stride = params.final_conv_stride
        norm_type = params.norm_type
        act_type = params.act_type
        self.return_stage_idx = params.return_stage_idx if params.return_stage_idx else [len(stage_params) - 1]
        self.layer_scale = params.layer_scale

        # Define model
        self.conv_stem = ConvLayer(3, stem_out_channel, kernel_size=stem_kernel_size, stride=stem_stride,
                                   bias=False, norm_type=norm_type, act_type=act_type)

        stages = []

        prev_channel = stem_out_channel
        for stage_param in stage_params:
            stage = []

            for i, block_info in enumerate(stage_param):
                block_type = block_info[0]
                if block_type == 'conv':
                    in_channels = prev_channel
                    out_channels = block_info[1]
                    kernel_size = block_info[2]
                    stride = block_info[3]

                    stage.append(
                        ConvLayer(in_channels, out_channels, kernel_size=kernel_size, stride=stride, 
                                  bias=False, norm_type=norm_type, act_type=act_type)
                    )

                elif block_type == 'fi':
                    in_channels = prev_channel
                    out_channels = block_info[1]
                    hidden_channels = block_info[2]
                    kernel_size = block_info[3]
                    stride = block_info[4]

                    stage.append(
                        FusedIB(in_channels, hidden_channels, out_channels, kernel_size, stride, norm_type, act_type)
                    )

                elif block_type == 'uir':
                    in_channels = prev_channel
                    out_channels = block_info[1]
                    hidden_channels = block_info[2]
                    extra_dw = block_info[3]
                    extra_kernel_size = block_info[4]
                    middle_dw = block_info[5]
                    middle_kernel_size = block_info[6]
                    stride = block_info[7]

                    stage.append(
                        UniversalInvertedResidualBlock(in_channels, hidden_channels, out_channels, extra_dw, extra_kernel_size, 
                                                       middle_dw, middle_kernel_size, stride, norm_type, act_type, layer_scale=self.layer_scale)
                    )

                elif block_type == 'mmqa':
                    in_channels = prev_channel
                    out_channels = block_info[1]
                    attention_channel = block_info[2]
                    num_attention_heads = block_info[3]
                    quary_pooling_stride = block_info[4]
                    key_val_downsample = block_info[5]
                    key_val_downsample_kernel_size = block_info[6]
                    key_val_downsample_stride = block_info[7]
                    stride = block_info[8]

                    assert in_channels == out_channels, 'MobileMultiQueryAttention2D requires in_channels == out_channels'
                    assert stride == 1, 'MobileMultiQueryAttention2D only supports stride=1'

                    stage.append(
                        MobileMultiQueryAttention2D(in_channels, num_attention_heads, attention_channel,
                                                    use_qkv_bias=False,
                                                    query_pooling_stride=quary_pooling_stride,
                                                    key_val_downsample=key_val_downsample,
                                                    key_val_downsample_kernel_size=key_val_downsample_kernel_size, 
                                                    key_val_downsample_stride=key_val_downsample_stride,
                                                    layer_scale=self.layer_scale)
                    )

                else:
                    raise ValueError(f'Unknown block type: {block_type}')

                prev_channel = out_channels

            stages.append(stage)

        # Add conv on last stage
        final_conv_in_channel = prev_channel
        stages[-1].append(
            ConvLayer(final_conv_in_channel, final_conv_out_channel, kernel_size=final_conv_kernel_size, 
                      stride=final_conv_stride, bias=False, norm_type=norm_type, act_type=act_type)
        )

        # Build stages
        stages = [nn.Sequential(*stage) for stage in stages]
        self.stages = nn.ModuleList(stages)

        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self._feature_dim = final_conv_out_channel
        stage_out_channels = [stage_param[-1][1] for stage_param in stage_params]
        stage_out_channels[-1] = final_conv_out_channel # Replace with final conv out channel
        self._intermediate_features_dim = [stage_out_channels[i] for i in self.return_stage_idx]

    def forward(self, x: Tensor):
        x = self.conv_stem(x)

        all_hidden_states = () if self.use_intermediate_features else None
        for stage_idx, stage in enumerate(self.stages):
            x = stage(x)
            if self.use_intermediate_features and stage_idx in self.return_stage_idx:
                all_hidden_states = all_hidden_states + (x, )

        if self.use_intermediate_features:
            return BackboneOutput(intermediate_features=all_hidden_states)
        
        x = self.avgpool(x)
        x = torch.flatten(x, 1)

        return BackboneOutput(last_feature=x)
    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def task_support(self, task):
        return task.lower() in SUPPORTING_TASK


def mobilenetv4(task, conf_model_backbone) -> MobileNetV4:
    return MobileNetV4(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path

from setuptools import find_packages, setup

version = (Path("src/netspresso_trainer") / "VERSION").read_text().strip()

readme_contents = Path("README.md").read_text()

requirements = Path("requirements.txt").read_text().split('\n')
requirements_all = requirements

setup(
    name="netspresso_trainer",
    version=version,
    author="NetsPresso",
    author_email="netspresso@nota.ai",
    description="NetsPresso Python Package",
    long_description=readme_contents,
    long_description_content_type="text/markdown",
    url="https://github.com/Nota-NetsPresso/netspresso-trainer",
    install_requires=requirements,
    package_dir={"": "src"},
    packages=find_packages("src", exclude=("tests",)),
    package_data={"": ["VERSION"]},
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "netspresso-train = netspresso_trainer.trainer_main:train_cli",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from loguru import logger

LOG_FILENAME = "result.log"

class StdOutLogger:
    def __init__(self, task, model, total_epochs=None, result_dir=None) -> None:
        super(StdOutLogger, self).__init__()
        self.task = task
        self.model_name = model
        self.total_epochs = total_epochs if total_epochs is not None else "???"

    def __call__(
        self,
        prefix: Literal['training', 'validation', 'evaluation', 'inference'],
        epoch: Optional[int] = None,
        losses : Optional[Dict] = None,
        metrics: Optional[Dict] = None,
        learning_rate: Optional[float] = None,
        elapsed_time: Optional[float] = None,
        **kwargs
    ):
        if epoch is not None and prefix == 'training':
            logger.info(f"Epoch: {epoch} / {self.total_epochs}")

        if learning_rate is not None:
            logger.info(f"learning rate: {learning_rate:.7f}")
        if elapsed_time is not None:
            logger.info(f"elapsed_time: {elapsed_time:.7f}")

        if losses is not None:
            logger.info(f"{prefix} loss: {losses['total']:.7f}")
        if metrics is not None:
            logger.info(f"{prefix} metric: {[(name, value) for name, value in metrics.items()]}")



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from itertools import chain
from typing import Any, Dict, Union

import torch
import torch.nn as nn

from ..utils.record import AverageMeter
from .registry import LOSS_DICT, PHASE_LIST


class LossFactory:
    def __init__(self, conf_model, **kwargs) -> None:
        self.loss_func_dict = {}
        self.loss_weight_dict = {}

        self.conf_model = conf_model
        self._build_losses(**kwargs)

        self._dirty_backward: bool = False
        self.loss_val_per_epoch: Dict[str, Dict[str, AverageMeter]] = {}
        self._clear_epoch_start()

    def _build_losses(self, **kwargs):

        for loss_element in self.conf_model.losses:
            criterion = loss_element.criterion
            loss_config = {k: v for k, v in loss_element.items() if k not in ['criterion', 'weight']}
            loss = LOSS_DICT[criterion](**loss_config, **kwargs)

            self.loss_func_dict.update({criterion: loss})
            loss_weight = loss_element.weight if loss_element.weight is not None else 1.0
            self.loss_weight_dict.update({criterion: loss_weight})

    def _clear_step_start(self):
        self.total_loss_for_backward: Union[int, torch.Tensor] = 0
        self._dirty_backward = False

    def _clear_epoch_start(self):
        self.loss_val_per_epoch = {
            phase: {
                loss_key: AverageMeter(loss_key, ':.4e') for loss_key in chain(self.loss_func_dict.keys(), ['total'])
            }
            for phase in PHASE_LIST
        }
        self._clear_step_start()

    def _assert_argument(self, kwargs):
        pass

    def backward(self, grad_scaler: torch.cuda.amp.GradScaler):
        assert not self._dirty_backward
        self.total_loss_for_backward.requires_grad_(True)
        grad_scaler.scale(self.total_loss_for_backward).backward()
        self._dirty_backward = True

    def result(self, phase='train'):
        phase = phase.lower()
        return self.loss_val_per_epoch[phase]

    def reset_values(self):
        self._clear_epoch_start()

    def calc(self, out: Dict, target: Union[torch.Tensor, Dict[str, torch.Tensor]], phase='train', **kwargs: Any) -> None:
        self.__call__(out=out, target=target, phase=phase, **kwargs)

    def __call__(self, out: Dict, target: Union[torch.Tensor, Dict[str, torch.Tensor]], phase: str, **kwargs: Any) -> None:
        phase = phase.lower()
        assert phase in PHASE_LIST, f"{phase} is not defined at our phase list ({PHASE_LIST})"

        self._assert_argument(kwargs)
        self._clear_step_start()

        for loss_key, loss_func in self.loss_func_dict.items():
            loss_val = loss_func(out, target)
            self.loss_val_per_epoch[phase][loss_key].update(loss_val.item())
            self.total_loss_for_backward += loss_val * self.loss_weight_dict[loss_key]

        self.loss_val_per_epoch[phase]['total'].update(self.total_loss_for_backward.item())


def build_losses(conf_model, **kwargs):
    loss_handler = LossFactory(conf_model, **kwargs)
    return loss_handler



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

# import logging
import sys
import time
from pathlib import Path
from typing import Literal, Optional, Union

import torch
import torch.distributed as dist
from loguru import logger
from omegaconf import DictConfig, ListConfig, OmegaConf

__all__ = ['set_logger', 'yaml_for_logging']

OUTPUT_ROOT_DIR = "./outputs"
LEVELNO_TO_LEVEL_NAME = {
    0: "NOTSET",
    10: "DEBUG",
    20: "INFO",
    30: "WARNING",
    40: "ERROR",
    50: "CRITICAL",
}

def rank_filter(record):
    try:
        return dist.get_rank() == 0
    except RuntimeError:  # Default process group has not been initialized, please make sure to call init_process_group.
        return True

def get_format(level: str, distributed: bool = False):
    debug_and_multi_gpu = (level == 'DEBUG' and distributed)
    fmt = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"

    if debug_and_multi_gpu:
        fmt = f"[GPU:{dist.get_rank()}] " + fmt

    only_rank_zero = not debug_and_multi_gpu
    return fmt, only_rank_zero

def add_stream_handler(level: str, distributed: bool):
    fmt, only_rank_zero = get_format(level, distributed=distributed)
    logger.add(sys.stderr, level=level, format=fmt, filter=rank_filter if only_rank_zero else "")



def add_file_handler(log_filepath: str, distributed: bool):
    level = LEVELNO_TO_LEVEL_NAME[logger._core.min_level]
    fmt, only_rank_zero = get_format(level, distributed=distributed)
    logger.add(log_filepath, level=level, format=fmt, filter=rank_filter if only_rank_zero else "", enqueue=True)



def _custom_logger(level: str, distributed: bool):
    logger.remove()
    add_stream_handler(level, distributed)

    return logger



def set_logger(level: str = 'INFO', distributed=False):
    try:
        time.tzset()
    except AttributeError as e:
        print(e)
        print("Skipping timezone setting.")
    _level: Literal['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'] = level.upper()
    _custom_logger(_level, distributed)

    return logger


def _yaml_for_logging(config: DictConfig) -> DictConfig:
    # TODO: better configuration logging
    list_maximum_index = 2
    new_config = OmegaConf.create()
    for k, v in config.items():
        if isinstance(v, DictConfig):
            new_config.update({k: _yaml_for_logging(v)})
        elif isinstance(v, ListConfig):
            new_config.update({k: list(map(str, v[:list_maximum_index])) + ['...']})
        else:
            new_config.update({k: v})
    return new_config


def yaml_for_logging(config: DictConfig):
    config_summarized = OmegaConf.create(_yaml_for_logging(config))
    return OmegaConf.to_yaml(config_summarized)


if __name__ == '__main__':
    set_logger(level='DEBUG')
